import typing

import keras

from keras_aug._src.keras_aug_export import keras_aug_export
from keras_aug._src.layers.base.vision_random_layer import VisionRandomLayer
from keras_aug._src.utils.argument_validation import standardize_bbox_format
from keras_aug._src.utils.argument_validation import standardize_data_format
from keras_aug._src.utils.argument_validation import standardize_padding_mode
from keras_aug._src.utils.argument_validation import standardize_size


@keras_aug_export(parent_path=["keras_aug.layers.vision"])
@keras.saving.register_keras_serializable(package="keras_aug")
class CenterCrop(VisionRandomLayer):
    """Crop the inputs at the center.

    If image size is smaller than output size along any edge, image is padded
    with `padding_value` and then center cropped.

    Args:
        size: Desired output size of the crop. If `size` is an int instead of
            sequence like `(h, w)`, a square crop `(size, size)` is made.
        padding_mode: The mode of the padding. Available values:
            `"constant", "edge", "reflect", "symmetric"`. Defaults to
            `"constant"`.
        padding_value: The padding value. It only affects when
            `padding_mode="constant"`. Defaults to `0`.
        bounding_box_format: The format of the bounding boxes. If specified,
            the available values are `"xyxy", "xywh", "center_xywh", "rel_xyxy",
            "rel_xywh", "rel_center_xywh"`. Defaults to `None`.
        data_format: A string specifying the data format of the input images.
            It can be either `"channels_last"` or `"channels_first"`.
            If not specified, the value will be interpreted by
            `keras.config.image_data_format`. Defaults to `None`.
    """

    def __init__(
        self,
        size: typing.Union[typing.Sequence[int], int],
        padding_mode: str = "constant",
        padding_value: float = 0,
        bounding_box_format: typing.Optional[str] = None,
        data_format: typing.Optional[str] = None,
        **kwargs,
    ):
        super().__init__(has_generator=False, **kwargs)
        self.size = standardize_size(size)
        self.padding_mode = standardize_padding_mode(padding_mode)
        self.padding_value = padding_value
        self.bounding_box_format = standardize_bbox_format(bounding_box_format)
        self.data_format = standardize_data_format(data_format)

        if self.data_format == "channels_last":
            self.h_axis, self.w_axis = -3, -2
        else:
            self.h_axis, self.w_axis = -2, -1

    def compute_output_shape(self, input_shape):
        images_shape, segmentation_masks_shape = self._get_shape_or_spec(
            input_shape
        )
        images_shape = list(images_shape)
        images_shape[self.h_axis] = self.size[0]
        images_shape[self.w_axis] = self.size[1]
        if segmentation_masks_shape is not None:
            segmentation_masks_shape = list(segmentation_masks_shape)
            segmentation_masks_shape[self.h_axis] = self.size[0]
            segmentation_masks_shape[self.w_axis] = self.size[1]
        return self._set_shape(
            input_shape, images_shape, segmentation_masks_shape
        )

    def get_params(self, batch_size, images=None, **kwargs):
        ops = self.backend
        images_shape = ops.shape(images)
        height, width = images_shape[self.h_axis], images_shape[self.w_axis]

        top = ops.numpy.where(
            height < self.size[0], (self.size[0] - height) / 2, 0
        )
        top = ops.cast(ops.numpy.round(top), "int32")
        bottom = ops.numpy.where(
            height < self.size[0], self.size[0] - height - top, 0
        )
        left = ops.numpy.where(
            width < self.size[1], (self.size[1] - width) / 2, 0
        )
        left = ops.cast(ops.numpy.round(left), "int32")
        right = ops.numpy.where(
            width < self.size[1], self.size[1] - width - left, 0
        )
        return dict(top=top, bottom=bottom, left=left, right=right)

    def augment_images(self, images, transformations, **kwargs):
        ops = self.backend
        images_shape = ops.shape(images)
        ori_height = images_shape[self.h_axis]
        ori_width = images_shape[self.w_axis]

        # Pad
        top = transformations["top"]
        bottom = transformations["bottom"]
        left = transformations["left"]
        right = transformations["right"]
        images = self.image_backend.pad(
            images,
            self.padding_mode,
            top,
            bottom,
            left,
            right,
            self.padding_value,
            self.data_format,
        )

        # Center crop
        offset_height = ops.numpy.floor_divide(
            (ori_height + top + bottom - self.size[0]), 2
        )
        offset_width = ops.numpy.floor_divide(
            (ori_width + left + right - self.size[1]), 2
        )
        images = self.image_backend.crop(
            images, offset_height, offset_width, self.size[0], self.size[1]
        )
        return ops.cast(images, self.image_dtype)

    def augment_labels(self, labels, transformations, **kwargs):
        return labels

    def augment_bounding_boxes(
        self,
        bounding_boxes,
        transformations,
        images=None,
        raw_images=None,
        **kwargs,
    ):
        if self.bounding_box_format is None:
            raise ValueError(
                f"{self.__class__.__name__} was called with bounding boxes,"
                "but no `bounding_box_format` was specified in the constructor."
            )
        ops = self.backend
        images_shape = ops.shape(raw_images)
        height = images_shape[self.h_axis]
        width = images_shape[self.w_axis]
        bounding_boxes = self.bbox_backend.convert_format(
            bounding_boxes,
            source=self.bounding_box_format,
            target="xyxy",
            height=height,
            width=width,
            dtype=self.bounding_box_dtype,
        )

        x1, y1, x2, y2 = ops.numpy.split(bounding_boxes["boxes"], 4, axis=-1)

        # Get pad and offset
        pad_top = ops.cast(transformations["top"], dtype="float32")
        pad_left = ops.cast(transformations["left"], dtype="float32")
        pad_bottom = ops.cast(transformations["bottom"], dtype="float32")
        pad_right = ops.cast(transformations["right"], dtype="float32")
        offset_height = ops.numpy.floor_divide(
            (height + pad_top + pad_bottom - self.size[0]), 2
        )
        offset_width = ops.numpy.floor_divide(
            (width + pad_left + pad_right - self.size[1]), 2
        )
        boxes = self.bbox_backend.pad(
            bounding_boxes["boxes"], pad_top, pad_left
        )
        boxes = self.bbox_backend.crop(
            boxes,
            offset_height,
            offset_width,
            height=self.size[0],
            width=self.size[1],
        )

        bounding_boxes = bounding_boxes.copy()
        bounding_boxes["boxes"] = boxes
        bounding_boxes = self.bbox_backend.clip_to_images(
            bounding_boxes,
            height=self.size[0],
            width=self.size[1],
            format="xyxy",
        )
        bounding_boxes = self.bbox_backend.convert_format(
            bounding_boxes,
            source="xyxy",
            target=self.bounding_box_format,
            height=self.size[0],
            width=self.size[1],
            dtype=self.bounding_box_dtype,
        )
        return bounding_boxes

    def augment_segmentation_masks(
        self, segmentation_masks, transformations, **kwargs
    ):
        ops = self.backend
        images_shape = ops.shape(segmentation_masks)
        ori_height = images_shape[self.h_axis]
        ori_width = images_shape[self.w_axis]

        # Pad
        top = transformations["top"]
        bottom = transformations["bottom"]
        left = transformations["left"]
        right = transformations["right"]
        segmentation_masks = self.image_backend.pad(
            segmentation_masks,
            "constant",
            top,
            bottom,
            left,
            right,
            -1,
            self.data_format,
        )

        # Center crop
        offset_height = ops.numpy.floor_divide(
            (ori_height + top + bottom - self.size[0]), 2
        )
        offset_width = ops.numpy.floor_divide(
            (ori_width + left + right - self.size[1]), 2
        )
        segmentation_masks = self.image_backend.crop(
            segmentation_masks,
            offset_height,
            offset_width,
            self.size[0],
            self.size[1],
        )
        return segmentation_masks

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "size": self.size,
                "padding_mode": self.padding_mode,
                "padding_value": self.padding_value,
                "bounding_box_format": self.bounding_box_format,
            }
        )
        return config



import keras
from keras import backend
from keras import tree
from keras.src.utils.backend_utils import in_tf_graph

from keras_aug._src.backend.bounding_box import BoundingBoxBackend
from keras_aug._src.backend.dynamic_backend import DynamicBackend
from keras_aug._src.backend.dynamic_backend import DynamicRandomGenerator
from keras_aug._src.backend.image import ImageBackend
from keras_aug._src.keras_aug_export import keras_aug_export


@keras_aug_export(parent_path=["keras_aug.layers.base"])
@keras.saving.register_keras_serializable(package="keras_aug")
class VisionRandomLayer(keras.Layer):
    """Abstract base layer for vectorized image augmentation.

    This layer contains base functionalities for preprocessing layers which
    augment image related data, e.g. image and in the future, label and bounding
    boxes. The subclasses could avoid making certain mistakes and reduce code
    duplications.

    This layer requires you to implement one method `augment_images()`, which
    augments one single image during the training. There are a few additional
    methods that you can implement for added functionality on the layer.

    `augment_labels()`, which handles label augmentation if the layer
    supports that.

    `augment_bounding_boxes()`, which handles the bounding box
    augmentation, if the layer supports that.

    `augment_keypoints()`, which handles the keypoints augmentation, if the
    layer supports that.

    `augment_segmentation_masks()`, which handles the segmentation masks
    augmentation, if the layer supports that.

    `augment_custom_annotations()`, which handles the custom annotations
    augmentation, if the layer supports that. This is useful to implement
    augmentation for special annotatinos.

    `get_random_transformations()`, which should produce a batch of random
    transformation settings. The transformation object, which must be a
    batched Tensor or a dictionary where each input is a batched Tensor,
    will be passed to `augment_images`, `augment_labels` and
    `augment_bounding_boxes`, to coordinate the randomness behavior, eg, in
    the RandomFlip layer, the image and bounding_boxes should be changed in
    the same way.

    The `call()` method support two formats of inputs::

        1. Single image tensor with 3D or 4D format.
        2. A dict of tensors with stable keys. The supported keys are
            `"images"`, `"labels"`, `"bounding_boxes"`,
            `segmentation_masks`, `keypoints` and `custom_annotations` at
            the moment. We might add more keys in future when we support more
            types of augmentation.

    The output of the `call()` will be in two formats, which will be the same
    structure as the inputs.

    The `call()` will unpack the inputs, forward to the correct function, and
    pack the output back to the same structure as the inputs.
    """

    IMAGES = "images"
    LABELS = "labels"
    BOUNDING_BOXES = "bounding_boxes"
    KEYPOINTS = "keypoints"
    SEGMENTATION_MASKS = "segmentation_masks"
    CUSTOM_ANNOTATIONS = "custom_annotations"

    IS_DICT = "is_dict"
    BATCHED = "batched"

    def __init__(self, has_generator=True, seed=None, **kwargs):
        super().__init__(**kwargs)
        # Check dtype
        if not backend.is_float_dtype(self.compute_dtype):
            if self.compute_dtype != "uint8":
                raise ValueError(
                    "Only floating and 'uint8' are supported for compute dtype."
                    f" Received: compute_dtype={self.compute_dtype}"
                )

        self._backend = DynamicBackend(backend.backend())
        if has_generator:
            self._random_generator = DynamicRandomGenerator(
                backend.backend(), seed=seed
            )
        self.has_generator = has_generator
        self.seed = seed

        # Other vision backends
        self.image_backend = ImageBackend(backend.backend())
        self.bbox_backend = BoundingBoxBackend(backend.backend())

        self._convert_input_args = False
        self._allow_non_tensor_positional_args = True
        self.autocast = False

    @property
    def image_dtype(self):
        return self.compute_dtype

    @property
    def bounding_box_dtype(self):
        dtype = backend.result_type(self.compute_dtype, float)
        return dtype

    @property
    def keypoint_dtype(self):
        dtype = backend.result_type(self.compute_dtype, float)
        return dtype

    @property
    def backend(self):
        return self._backend.backend

    @property
    def random_generator(self):
        return self._random_generator.random_generator

    def get_params(
        self,
        batch_size,
        images=None,
        labels=None,
        bounding_boxes=None,
        keypoints=None,
        segmentation_masks=None,
        custom_annotations=None,
    ):
        """Produce transformations parameters.

        Returns:
            Any type of object, which will be forwarded as the
            `transformations` parameter.
        """
        return self.backend.numpy.zeros([batch_size])

    def augment_images(self, images, transformations, **kwargs):
        """Augment a batch of images.

        Args:
            images: 4D image input tensor to the layer. Forwarded from
                `layer.call()`. This should generally have the shape
                [B, H, W, C]. Forwarded from `layer.call()`.
            transformations: The transformations object produced by
                `get_random_transformations`. Used to coordinate the randomness
                between image, label, bounding box, keypoints, and segmentation
                mask.

        Returns:
            output 4D tensor, which will be forward to `layer.call()`.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} doesn't support augmenting images."
        )

    def augment_labels(self, labels, transformations, **kwargs):
        """Augment a batch of labels.

        Args:
            labels: 2D label to the layer. Forwarded from `layer.call()`.
            transformations: The transformations object produced by
                `get_random_transformations`. Used to coordinate the randomness
                between image, label, bounding box, keypoints, and segmentation
                mask.

        Returns:
            output 2D tensor, which will be forward to `layer.call()`.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} doesn't support augmenting labels."
        )

    def augment_bounding_boxes(self, bounding_boxes, transformations, **kwargs):
        """Augment bounding boxes for one image.

        Args:
            bounding_boxes: 3D bounding boxes to the layer. Forwarded from
                `call()`.
            transformations: The transformations object produced by
                `get_random_transformations`. Used to coordinate the randomness
                between image, label, bounding box, keypoints, and segmentation
                mask.

        Returns:
            output 3D tensor, which will be forward to `layer.call()`.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} doesn't support augmenting "
            "bounding boxes."
        )

    def augment_keypoints(self, keypoints, transformations, **kwargs):
        """Augment a batch of keypoints for one image.

        Args:
            keypoints: 3D keypoints input tensor to the layer. Forwarded from
                `layer.call()`. Shape should be [batch, num_keypoints, 2] in the
                specified keypoint format.
            transformations: The transformations object produced by
                `get_random_transformations`. Used to coordinate the randomness
                between image, label, bounding box, keypoints, and segmentation
                mask.

        Returns:
            output 3D tensor, which will be forward to `layer.call()`.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} doesn't support augmenting "
            "keypoints."
        )

    def augment_segmentation_masks(
        self, segmentation_masks, transformations, **kwargs
    ):
        """Augment a batch of images' segmentation masks.

        Args:
            segmentation_masks: 4D segmentation mask input tensor to the layer.
                This should generally have the shape [B, H, W, 1], or in some
                cases [B, H, W, C] for multilabeled data. Forwarded from
                `layer.call()`.
            transformations: The transformations object produced by
                `get_random_transformations`. Used to coordinate the randomness
                between image, label, bounding box, keypoints, and segmentation
                mask.

        Returns:
            output 4D tensor containing the augmented segmentation mask, which
            will be forward to `layer.call()`.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} doesn't support augmenting "
            "segmentation masks."
        )

    def augment_custom_annotations(
        self, custom_annotations, transformations, **kwargs
    ):
        """Augment a batch of images' custom_annotations.

        Args:
            custom_annotations: 4D custom annotations input tensor to the layer.
            transformations: The transformations object produced by
                `get_random_transformations`. Used to coordinate the randomness
                between image, label, bounding box, keypoints, and segmentation
                mask.

        Returns:
            output 4D tensor containing the augmented segmentation mask, which
            will be forward to `layer.call()`.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} doesn't support augmenting "
            "custom annotations."
        )

    def _batch_augment(self, inputs, **kwargs):
        images = inputs.get(self.IMAGES, None)
        raw_images = images
        labels = inputs.get(self.LABELS, None)
        bounding_boxes = inputs.get(self.BOUNDING_BOXES, None)
        keypoints = inputs.get(self.KEYPOINTS, None)
        segmentation_masks = inputs.get(self.SEGMENTATION_MASKS, None)
        custom_annotations = inputs.get(self.CUSTOM_ANNOTATIONS, None)
        batch_size = self.backend.shape(images)[0]

        transformations = self.get_params(
            batch_size,
            images=images,
            labels=labels,
            bounding_boxes=bounding_boxes,
            keypoints=keypoints,
            segmentation_masks=segmentation_masks,
            custom_annotations=custom_annotations,
        )

        images = self.augment_images(
            images,
            transformations=transformations,
            bounding_boxes=bounding_boxes,
            labels=labels,
        )
        result = {self.IMAGES: images}

        if labels is not None:
            labels = self.augment_labels(
                labels,
                transformations=transformations,
                bounding_boxes=bounding_boxes,
                images=images,
                raw_images=raw_images,
            )
            result[self.LABELS] = labels

        if bounding_boxes is not None:
            bounding_boxes = self.augment_bounding_boxes(
                bounding_boxes,
                transformations=transformations,
                labels=labels,
                images=images,
                raw_images=raw_images,
            )
            result[self.BOUNDING_BOXES] = bounding_boxes

        if keypoints is not None:
            keypoints = self.augment_keypoints(
                keypoints,
                transformations=transformations,
                labels=labels,
                bounding_boxes=bounding_boxes,
                images=images,
                raw_images=raw_images,
            )
            result[self.KEYPOINTS] = keypoints

        if segmentation_masks is not None:
            segmentation_masks = self.augment_segmentation_masks(
                segmentation_masks,
                transformations=transformations,
                labels=labels,
                bounding_boxes=bounding_boxes,
                images=images,
                raw_images=raw_images,
            )
            result[self.SEGMENTATION_MASKS] = segmentation_masks

        if custom_annotations is not None:
            custom_annotations = self.augment_custom_annotations(
                custom_annotations,
                transformations=transformations,
                labels=labels,
                bounding_boxes=bounding_boxes,
                images=images,
                raw_images=raw_images,
            )
            result[self.CUSTOM_ANNOTATIONS] = custom_annotations

        # preserve any additional inputs unmodified by this layer.
        for key in inputs.keys() - result.keys():
            result[key] = inputs[key]
        return result

    def _format_inputs(self, inputs):
        ops = self.backend
        metadata = {self.IS_DICT: True}
        if not isinstance(inputs, dict):
            metadata[self.IS_DICT] = False
            inputs = {self.IMAGES: inputs}
        if not isinstance(inputs, dict):
            raise ValueError(
                "Expect the inputs to be image tensor or dict. "
                f"Received: inputs={inputs} of type={type(inputs)}"
            )

        images_shape = ops.shape(inputs["images"])
        metadata[self.BATCHED] = len(images_shape) == 4
        if len(images_shape) == 3:
            for key in list(inputs.keys()):
                if key == self.BOUNDING_BOXES:
                    inputs[self.BOUNDING_BOXES]["boxes"] = (
                        ops.numpy.expand_dims(
                            inputs[self.BOUNDING_BOXES]["boxes"], axis=0
                        )
                    )
                    inputs[self.BOUNDING_BOXES]["classes"] = (
                        ops.numpy.expand_dims(
                            inputs[self.BOUNDING_BOXES]["classes"], axis=0
                        )
                    )
                else:
                    inputs[key] = ops.numpy.expand_dims(inputs[key], axis=0)
        if self.BOUNDING_BOXES in inputs:
            inputs[self.BOUNDING_BOXES] = self._format_bounding_boxes(
                inputs[self.BOUNDING_BOXES]
            )
        return inputs, metadata

    def _cast_inputs(self, inputs):
        if not isinstance(inputs, dict):
            raise TypeError
        ops = self.backend
        inputs = inputs.copy()
        if self.IMAGES in inputs:
            inputs[self.IMAGES] = ops.convert_to_tensor(inputs[self.IMAGES])
            inputs[self.IMAGES] = self.image_backend.transform_dtype(
                inputs[self.IMAGES], inputs[self.IMAGES].dtype, self.image_dtype
            )
        if self.LABELS in inputs:
            inputs[self.LABELS] = ops.convert_to_tensor(inputs[self.LABELS])
        if self.BOUNDING_BOXES in inputs:
            bounding_boxes = inputs[self.BOUNDING_BOXES].copy()
            bounding_boxes["boxes"] = ops.convert_to_tensor(
                bounding_boxes["boxes"], self.bounding_box_dtype
            )
            bounding_boxes["classes"] = ops.convert_to_tensor(
                bounding_boxes["classes"], self.bounding_box_dtype
            )
            inputs[self.BOUNDING_BOXES] = bounding_boxes
        if self.SEGMENTATION_MASKS in inputs:
            masks = inputs[self.SEGMENTATION_MASKS]
            masks = ops.convert_to_tensor(masks)
            if backend.is_float_dtype(masks.dtype) and backend.is_float_dtype(
                self.compute_dtype
            ):
                masks = ops.cast(masks, self.compute_dtype)
            inputs[self.SEGMENTATION_MASKS] = masks
        if self.KEYPOINTS in inputs:
            inputs[self.KEYPOINTS] = ops.convert_to_tensor(
                inputs[self.KEYPOINTS], self.keypoint_dtype
            )
        if self.CUSTOM_ANNOTATIONS in inputs:
            inputs[self.CUSTOM_ANNOTATIONS] = tree.map_structure(
                ops.convert_to_tensor, inputs[self.CUSTOM_ANNOTATIONS]
            )
        return inputs

    def _format_bounding_boxes(self, bounding_boxes):
        # We can't catch the case where this is None, sometimes RaggedTensor
        # drops this dimension.
        if "classes" not in bounding_boxes:
            raise ValueError(
                "Bounding boxes are missing class_id. If you would like to pad "
                "the bounding boxes with class_id, use: "
                "`bounding_boxes['classes'] = "
                "tf.ones_like(bounding_boxes['boxes'])`."
            )
        return bounding_boxes

    def _format_output(self, output, metadata):
        ops = self.backend
        if not metadata[self.BATCHED]:
            for key in list(output.keys()):
                if key == self.BOUNDING_BOXES:
                    output[self.BOUNDING_BOXES]["boxes"] = ops.numpy.squeeze(
                        output[self.BOUNDING_BOXES]["boxes"], axis=0
                    )
                    output[self.BOUNDING_BOXES]["classes"] = ops.numpy.squeeze(
                        output[self.BOUNDING_BOXES]["classes"], axis=0
                    )
                else:
                    output[key] = ops.numpy.squeeze(output[key], axis=0)
        if not metadata[self.IS_DICT]:
            return output[self.IMAGES]
        return output

    def __call__(self, inputs, **kwargs):
        if in_tf_graph():
            self._set_backend("tensorflow")
            try:
                outputs = super().__call__(inputs, **kwargs)
            finally:
                self._reset_backend()
            return outputs
        else:
            return super().__call__(inputs, **kwargs)

    def call(self, inputs, **kwargs):
        ops = self.backend
        inputs, metadata = self._format_inputs(inputs)
        inputs = self._cast_inputs(inputs)
        images = inputs[self.IMAGES]
        images_shape = ops.shape(images)
        if len(images_shape) == 4:
            return self._format_output(self._batch_augment(inputs), metadata)
        else:
            raise ValueError(
                "Image augmentation layers are expecting inputs to be "
                "rank 3D (unbatched) or 4D (batched) tensors. "
                f"Received: images.shape={images.shape}"
            )

    def get_config(self):
        config = super().get_config()
        config.update({"seed": self.seed})
        return config

    @classmethod
    def from_config(cls, config):
        return cls(**config)

    # Useful functions

    def transform_value_range(
        self, images, original_range, target_range, dtype="float32"
    ):
        ops = self.backend
        original_range = tuple(original_range)
        target_range = tuple(target_range)

        images = ops.cast(images, dtype)
        images = ops.numpy.divide(
            (images - original_range[0]),
            (original_range[1] - original_range[0]),
        )
        scale_factor = ops.numpy.subtract(target_range[1], target_range[0])
        images = ops.numpy.add(
            ops.numpy.multiply(images, scale_factor), target_range[0]
        )
        return images

    def _set_backend(self, name):
        self._backend.set_backend(name)
        self.image_backend.set_backend(name)
        self.bbox_backend.set_backend(name)
        if self.has_generator:
            self._random_generator.set_generator(name)

    def _reset_backend(self):
        self._backend.reset()
        self.image_backend.reset()
        self.bbox_backend.reset()
        if self.has_generator:
            self._random_generator.reset()

    def _get_shape_or_spec(self, input_shape_or_inputs):
        """Get the shape or spec of `images` and `segmentation_masks`."""
        if not isinstance(input_shape_or_inputs, dict):
            return input_shape_or_inputs, None
        else:
            images = input_shape_or_inputs[self.IMAGES]
            segmentation_masks = input_shape_or_inputs.get(
                self.SEGMENTATION_MASKS, None
            )
            return images, segmentation_masks

    def _set_shape(
        self, original_shape, images_shape, segmentation_masks_shape=None
    ):
        images_shape = tuple(images_shape)
        if segmentation_masks_shape is not None:
            segmentation_masks_shape = tuple(segmentation_masks_shape)
        if not isinstance(original_shape, dict):
            return images_shape
        else:
            original_shape[self.IMAGES] = images_shape
            if self.SEGMENTATION_MASKS in original_shape:
                original_shape[self.SEGMENTATION_MASKS] = (
                    segmentation_masks_shape
                )
            return original_shape

    def _set_spec(
        self, original_spec, images_spec, segmentation_masks_spec=None
    ):
        if not isinstance(original_spec, dict):
            return images_spec
        else:
            original_spec[self.IMAGES] = images_spec
            if self.SEGMENTATION_MASKS in original_spec:
                original_spec[self.SEGMENTATION_MASKS] = segmentation_masks_spec
            return original_spec



import keras
import numpy as np
from absl.testing import parameterized
from keras import backend
from keras.src import testing

from keras_aug._src.layers.composition.random_order import RandomOrder
from keras_aug._src.layers.vision.random_grayscale import RandomGrayscale
from keras_aug._src.layers.vision.random_invert import RandomInvert
from keras_aug._src.layers.vision.resize import Resize
from keras_aug._src.utils.test_utils import get_images


class FixedRandomOrder(RandomOrder):
    def get_params(self):
        ops = self.backend
        fn_idx = ops.convert_to_tensor([1, 0], dtype="int32")
        return fn_idx


class RandomOrderTest(testing.TestCase, parameterized.TestCase):
    def setUp(self):
        # Defaults to channels_last
        self.data_format = backend.image_data_format()
        backend.set_image_data_format("channels_last")
        return super().setUp()

    def tearDown(self) -> None:
        backend.set_image_data_format(self.data_format)
        return super().tearDown()

    def test_correctness(self):
        import torch
        import torchvision.transforms.v2.functional as TF

        layer = FixedRandomOrder(
            transforms=[RandomGrayscale(p=1.0), RandomInvert(p=1.0)]
        )

        x = get_images("float32", "channels_last")
        y = layer(x)

        ref_y = TF.rgb_to_grayscale(
            TF.invert(torch.tensor(np.transpose(x, [0, 3, 1, 2]))),
            num_output_channels=3,
        )
        ref_y = np.transpose(ref_y.cpu().numpy(), [0, 2, 3, 1])
        self.assertAllClose(y, ref_y)

    def test_shape(self):
        layer = RandomOrder(
            transforms=[RandomGrayscale(p=1.0), RandomInvert(p=1.0)]
        )

        # Test dynamic shape
        x = keras.KerasTensor((None, None, None, 3))
        y = layer(x)
        self.assertEqual(y.shape, (None, None, None, 3))

        # Test static shape
        x = keras.KerasTensor((None, 32, 32, 3))
        y = layer(x)
        self.assertEqual(y.shape, (None, 32, 32, 3))

        # Test deterministic shape
        layer = RandomOrder(
            transforms=[Resize((16, 16)), RandomGrayscale(p=1.0)]
        )
        x = keras.KerasTensor((None, 16, 16, 3))
        y = layer(x)
        self.assertEqual(y.shape, (None, 16, 16, 3))

    def test_model(self):
        layer = RandomOrder(
            transforms=[RandomGrayscale(p=1.0), RandomInvert(p=1.0)]
        )
        inputs = keras.layers.Input(shape=[None, None, 3])
        outputs = layer(inputs)
        model = keras.models.Model(inputs, outputs)
        self.assertEqual(model.output_shape, (None, None, None, 3))

    def test_config(self):
        x = get_images("float32", "channels_last")
        layer = FixedRandomOrder(
            transforms=[RandomGrayscale(p=1.0), RandomInvert(p=1.0)]
        )
        y = layer(x)

        layer = FixedRandomOrder.from_config(layer.get_config())
        y2 = layer(x)
        self.assertAllClose(y, y2)

    def test_tf_data_compatibility(self):
        import tensorflow as tf

        layer = RandomOrder(
            transforms=[RandomGrayscale(p=1.0), RandomInvert(p=1.0)]
        )
        x = get_images("float32", "channels_last")
        ds = tf.data.Dataset.from_tensor_slices(x).batch(2).map(layer)
        for output in ds.take(1):
            self.assertIsInstance(output, tf.Tensor)
            self.assertEqual(output.shape, (2, 32, 32, 3))



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from keras_aug._src.layers.composition.random_apply import RandomApply
from keras_aug._src.layers.composition.random_choice import RandomChoice
from keras_aug._src.layers.composition.random_order import RandomOrder



import math

from keras import backend

from keras_aug._src.backend.dynamic_backend import DynamicBackend
from keras_aug._src.utils.argument_validation import standardize_interpolation


class ImageBackend(DynamicBackend):
    def __init__(self, name=None):
        super().__init__(name=name)

    def transform_dtype(self, images, from_dtype, to_dtype):
        # Ref: torchvision.transforms.v2.ToDtype
        ops = self.backend
        from_dtype = backend.standardize_dtype(from_dtype)
        to_dtype = backend.standardize_dtype(to_dtype)

        if from_dtype == to_dtype:
            return images

        is_float_input = backend.is_float_dtype(from_dtype)
        is_float_output = backend.is_float_dtype(to_dtype)

        if is_float_input:
            # float to float
            if is_float_output:
                return ops.cast(images, to_dtype)

            # float to int
            if (from_dtype == "float32" and to_dtype in ("int32", "int64")) or (
                from_dtype == "float64" and to_dtype == "int64"
            ):
                raise ValueError(
                    f"The conversion from {from_dtype} to {to_dtype} cannot be "
                    "performed safely."
                )
            # See https://github.com/pytorch/vision/pull/2078#issuecomment-613524965 for details # noqa: E501
            eps = 1e-3
            max_value = float(self._max_value_of_dtype(to_dtype))
            return ops.cast(
                ops.numpy.multiply(images, max_value + 1.0 - eps), to_dtype
            )
        else:
            # int to float
            if is_float_output:
                max_value = float(self._max_value_of_dtype(from_dtype))
                return ops.numpy.divide(ops.cast(images, to_dtype), max_value)

            # int to int
            num_bits_input = self._num_bits_of_dtype(from_dtype)
            num_bits_output = self._num_bits_of_dtype(to_dtype)

            if num_bits_input > num_bits_output:
                return ops.cast(
                    images >> (num_bits_input - num_bits_output), to_dtype
                )
            else:
                return ops.cast(images, to_dtype) << (
                    num_bits_output - num_bits_input
                )

    def crop(self, images, top, left, height, width, data_format=None):
        data_format = data_format or backend.image_data_format()

        ops = self.backend
        images_shape = ops.shape(images)
        start_indices = [top, left]
        lengths = [height, width]
        if data_format == "channels_last":
            start_indices = start_indices + [0]
            lengths = lengths + [images_shape[-1]]
        else:
            start_indices = [0] + start_indices
            lengths = [images_shape[-3]] + lengths
        start_indices = [0] + start_indices
        lengths = [images_shape[0]] + lengths
        images = ops.core.slice(
            images, ops.cast(start_indices, "int32"), lengths
        )
        return images

    def pad(
        self,
        images,
        mode,
        top,
        bottom,
        left,
        right,
        constant_value=0,
        data_format=None,
    ):
        data_format = data_format or backend.image_data_format()

        ops = self.backend
        if self.name == "torch":  # Workaround for torch
            top = int(top)
            bottom = int(bottom)
            left = int(left)
            right = int(right)
        pad_width = [[top, bottom], [left, right]]
        if data_format == "channels_last":
            pad_width = pad_width + [[0, 0]]
        else:
            pad_width = [[0, 0]] + pad_width
        pad_width = [[0, 0]] + pad_width  # 4D

        images = ops.numpy.pad(
            images,
            pad_width,
            mode,
            constant_value if mode == "constant" else None,
        )
        return images

    def adjust_brightness(self, images, factor):
        ops = self.backend
        images = ops.convert_to_tensor(images)
        factor = ops.convert_to_tensor(factor)
        original_dtype = backend.standardize_dtype(images.dtype)
        is_float_inputs = backend.is_float_dtype(original_dtype)
        max_value = self._max_value_of_dtype(original_dtype)
        if len(ops.shape(factor)) == 1:
            factor = ops.numpy.expand_dims(factor, [1, 2, 3])

        images = ops.numpy.multiply(images, factor)
        images = ops.numpy.clip(images, 0, max_value)
        if not is_float_inputs:
            images = ops.cast(images, original_dtype)
        return images

    def adjust_contrast(self, images, factor, data_format=None):
        data_format = data_format or backend.image_data_format()

        ops = self.backend
        images = ops.convert_to_tensor(images)
        factor = ops.convert_to_tensor(factor)
        original_dtype = backend.standardize_dtype(images.dtype)
        is_float_inputs = backend.is_float_dtype(original_dtype)
        if len(ops.shape(factor)) == 1:
            factor = ops.numpy.expand_dims(factor, [1, 2, 3])

        grayscales = ops.image.rgb_to_grayscale(images, data_format)
        if not is_float_inputs:
            grayscales = ops.numpy.floor(grayscales)
        means = ops.numpy.mean(grayscales, axis=[-3, -2, -1], keepdims=True)
        images = self.blend(images, means, factor)
        return images

    def adjust_saturation(self, images, factor, data_format=None):
        data_format = data_format or backend.image_data_format()

        ops = self.backend
        images = ops.convert_to_tensor(images)
        factor = ops.convert_to_tensor(factor)
        original_dtype = backend.standardize_dtype(images.dtype)
        is_float_inputs = backend.is_float_dtype(original_dtype)
        if len(ops.shape(factor)) == 1:
            factor = ops.numpy.expand_dims(factor, [1, 2, 3])

        grayscales = ops.image.rgb_to_grayscale(images, data_format)
        if not is_float_inputs:
            grayscales = ops.numpy.floor(grayscales)
        images = self.blend(images, grayscales, factor)
        return images

    def adjust_hue(self, images, factor, data_format=None):
        data_format = data_format or backend.image_data_format()
        channels_axis = -1 if data_format == "channels_last" else -3

        ops = self.backend
        images = ops.convert_to_tensor(images)
        factor = ops.convert_to_tensor(factor)
        original_dtype = backend.standardize_dtype(images.dtype)
        max_value = self._max_value_of_dtype(original_dtype)
        if len(ops.shape(factor)) == 1:
            factor = ops.numpy.expand_dims(factor, [1, 2, 3])

        images = self.transform_dtype(
            images, images.dtype, backend.result_type(original_dtype, float)
        )
        images = ops.image.rgb_to_hsv(images, data_format)
        h, s, v = ops.numpy.split(images, 3, channels_axis)
        h = ops.numpy.add(h, factor)
        h = ops.numpy.mod(h, 1.0)
        images = ops.numpy.concatenate([h, s, v], channels_axis)
        images = ops.image.hsv_to_rgb(images, data_format)
        images = ops.numpy.clip(images, 0, max_value)
        images = self.transform_dtype(images, images.dtype, original_dtype)
        return images

    def affine(
        self,
        images,
        angle,
        translate_x,
        translate_y,
        scale,
        shear_x,
        shear_y,
        center_x=None,
        center_y=None,
        interpolation="bilinear",
        padding_mode="constant",
        padding_value=0,
        data_format=None,
    ):
        interpolation = standardize_interpolation(interpolation)
        data_format = data_format or backend.image_data_format()
        if data_format == "channels_last":
            h_axis, w_axis = -3, -2
        else:
            h_axis, w_axis = -2, -1

        ops = self.backend
        images = ops.convert_to_tensor(images)
        original_dtype = backend.standardize_dtype(images.dtype)
        images_shape = ops.shape(images)
        height, width = images_shape[h_axis], images_shape[w_axis]
        if center_x is None:
            center_x = 0.5
        if center_y is None:
            center_y = 0.5
        matrix = self._compute_affine_matrix(
            center_x,
            center_y,
            angle,
            translate_x,
            translate_y,
            scale,
            shear_x,
            shear_y,
            height,
            width,
        )
        transform = ops.numpy.reshape(matrix, [-1, 9])[:, :8]
        images = self.transform_dtype(
            images, original_dtype, backend.result_type(original_dtype, float)
        )
        # Fix padding_value
        if not backend.is_float_dtype(original_dtype):
            if original_dtype == "uint8":
                padding_value = float(padding_value) / 255.0
            else:
                raise NotImplementedError(
                    f"Unrecognize images.dtype={original_dtype}. Supported "
                    "dtypes are floating and 'uint8'."
                )
        images = ops.image.affine_transform(
            images,
            transform,
            interpolation,
            padding_mode,
            padding_value,
            data_format,
        )
        images = self.transform_dtype(images, images.dtype, original_dtype)
        return images

    def auto_contrast(self, images, data_format=None):
        data_format = data_format or backend.image_data_format()
        axis = (1, 2) if data_format == "channels_last" else (2, 3)

        ops = self.backend
        original_dtype = backend.standardize_dtype(images.dtype)
        max_value = self._max_value_of_dtype(original_dtype)
        if not backend.is_float_dtype(original_dtype):
            images = ops.cast(
                images, backend.result_type(original_dtype, float)
            )

        lows = ops.numpy.min(images, axis=axis, keepdims=True)
        highs = ops.numpy.max(images, axis=axis, keepdims=True)
        eq_index = ops.numpy.equal(lows, highs)
        inverse_scale = ops.numpy.divide(
            ops.numpy.subtract(highs, lows), max_value
        )
        lows = ops.numpy.where(eq_index, 0.0, lows)
        inverse_scale = ops.numpy.where(eq_index, 1.0, inverse_scale)

        images = ops.numpy.divide(
            ops.numpy.subtract(images, lows), inverse_scale
        )
        images = ops.numpy.clip(images, 0, max_value)
        images = ops.cast(images, original_dtype)
        return images

    def blend(self, images1, images2, factor):
        ops = self.backend
        images1 = ops.convert_to_tensor(images1)
        images2 = ops.convert_to_tensor(images2)
        original_dtype = backend.standardize_dtype(images1.dtype)
        is_float_inputs = backend.is_float_dtype(original_dtype)
        max_value = self._max_value_of_dtype(original_dtype)

        images1 = ops.numpy.multiply(images1, factor)
        images2 = ops.numpy.multiply(images2, (1.0 - factor))
        images = ops.numpy.add(images1, images2)
        images = ops.numpy.clip(images, 0, max_value)
        if not is_float_inputs:
            images = ops.cast(images, original_dtype)
        return images

    def equalize(self, images, bins=256, data_format=None):
        if bins != 256:
            raise NotImplementedError("`bins` must be `256`.")
        data_format = data_format or backend.image_data_format()

        ops = self.backend
        images = ops.convert_to_tensor(images)
        original_dtype = backend.standardize_dtype(images.dtype)
        images_shape = ops.shape(images)
        images = self.transform_dtype(images, images.dtype, "uint8")

        def _scale_channel(images):
            images = ops.cast(images, "int32")
            if self.name == "tensorflow":
                import tensorflow as tf

                hist = tf.histogram_fixed_width(
                    tf.reshape(images, [-1]), [0, 255], nbins=256
                )
            elif self.name == "jax":
                import jax.numpy as jnp

                hist = jnp.bincount(
                    jnp.reshape(images, [-1]), minlength=bins, length=bins
                )
            else:
                hist = ops.numpy.bincount(
                    ops.numpy.reshape(images, [-1]), minlength=bins
                )

            big_number = 1410065408
            nonzero_hist = ops.numpy.where(
                ops.numpy.equal(hist, 0), big_number, hist
            )
            step = ops.numpy.floor_divide(
                ops.numpy.sum(hist) - ops.numpy.min(nonzero_hist), 255
            )

            def build_mapping(histogram, step):
                # Replace where step is 0 with 1 to avoid division by 0.
                # This doesn't change the result, because where step==0 the
                # original image is returned
                _step = ops.numpy.where(ops.numpy.equal(step, 0), 1, step)

                # Compute the cumulative sum, shifting by step // 2
                # and then normalization by step.
                lookup_table = ops.numpy.floor_divide(
                    ops.numpy.add(
                        ops.numpy.cumsum(histogram),
                        ops.numpy.floor_divide(_step, 2),
                    ),
                    _step,
                )

                # Shift lookup_table, prepending with 0.
                lookup_table = ops.numpy.pad(lookup_table[:-1], [[1, 0]])

                # Clip the counts to be in range. This is done
                # in the C code for image.point.
                return ops.numpy.clip(lookup_table, 0, 255)

            def step_is_0():
                return ops.cast(images, "uint8")

            def step_not_0():
                lut = build_mapping(hist, step)
                result = ops.numpy.take(lut, images)
                return ops.cast(result, "uint8")

            return ops.cond(step == 0, step_is_0, step_not_0)

        def _equalize_single_image(image):
            if data_format == "channels_last":
                return ops.numpy.stack(
                    [
                        _scale_channel(image[..., c])
                        for c in range(image.shape[-1])
                    ],
                    axis=-1,
                )
            else:
                return ops.numpy.stack(
                    [_scale_channel(image[c]) for c in range(image.shape[-3])],
                    axis=-3,
                )

        # Workaround for tf.data
        if self.name == "tensorflow":
            import tensorflow as tf

            images = tf.map_fn(_equalize_single_image, images)
        else:
            images = ops.numpy.stack(
                [
                    _equalize_single_image(x)
                    for x in ops.core.unstack(images, axis=0)
                ],
                axis=0,
            )
        images = ops.numpy.reshape(images, images_shape)
        images = self.transform_dtype(images, images.dtype, original_dtype)
        return images

    def guassian_blur(self, images, kernel_size, sigma, data_format=None):
        data_format = data_format or backend.image_data_format()
        channels_axis = -1 if data_format == "channels_last" else -3

        ops = self.backend
        images = ops.core.convert_to_tensor(images)
        channels = ops.shape(images)[channels_axis]
        original_dtype = backend.standardize_dtype(images.dtype)
        compute_dtype = backend.result_type(original_dtype, float)

        def _get_gaussian_kernel_1d(kernel_size, sigma, dtype):
            lim = (kernel_size - 1) / (2.0 * math.sqrt(2.0) * sigma)
            x = ops.numpy.linspace(-lim, lim, num=kernel_size, dtype=dtype)
            kernel_1d = ops.nn.softmax(-ops.numpy.power(x, 2), axis=0)
            return kernel_1d

        def _get_gaussian_kernel_2d(kernel_size, sigma, dtype):
            kernel_1d_x = _get_gaussian_kernel_1d(
                kernel_size[0], sigma[0], dtype
            )
            kernel_1d_y = _get_gaussian_kernel_1d(
                kernel_size[1], sigma[1], dtype
            )
            kernel_2d = ops.numpy.multiply(
                ops.numpy.expand_dims(kernel_1d_y, axis=-1), kernel_1d_x
            )
            return kernel_2d

        images = ops.cast(images, compute_dtype)
        kernel = _get_gaussian_kernel_2d(kernel_size, sigma, compute_dtype)

        kernel = ops.numpy.expand_dims(kernel, axis=(2, 3))
        kernel = ops.numpy.tile(kernel, [1, 1, channels, 1])

        # Padding
        pad_width = [
            [kernel_size[1] // 2, kernel_size[1] // 2],  # y
            [kernel_size[0] // 2, kernel_size[0] // 2],  # x
        ]
        if data_format == "channels_last":
            pad_width = pad_width + [[0, 0]]
        else:
            pad_width = [[0, 0]] + pad_width
        pad_width = [[0, 0]] + pad_width
        images = ops.numpy.pad(images, pad_width, mode="reflect")
        images = ops.nn.depthwise_conv(images, kernel, data_format=data_format)
        if backend.is_int_dtype(original_dtype):
            images = ops.numpy.round(images)
            images = ops.cast(images, original_dtype)
        return images

    def rgb_to_grayscale(self, images, num_channels=3, data_format=None):
        if num_channels not in (1, 3):
            raise ValueError(
                "`num_channels` must be 1 or 3. "
                f"Received: num_channels={num_channels}"
            )
        data_format = data_format or backend.image_data_format()
        channels_axis = -1 if data_format == "channels_last" else -3

        ops = self.backend
        images = ops.core.convert_to_tensor(images)
        original_dtype = backend.standardize_dtype(images.dtype)
        images = ops.image.rgb_to_grayscale(images, data_format)
        images = ops.core.cast(images, original_dtype)
        if num_channels == 3:
            images = ops.numpy.repeat(images, 3, axis=channels_axis)
        return images

    def invert(self, images):
        ops = self.backend
        images = ops.convert_to_tensor(images)
        dtype = backend.standardize_dtype(images.dtype)
        if backend.is_float_dtype(dtype):
            return ops.numpy.subtract(1.0, images)
        elif "uint" in dtype:
            return ~images
        else:
            # signed integer dtypes
            num_bits = self._num_bits_of_dtype(dtype)
            return images ^ ((1 << num_bits) - 1)

    def posterize(self, images, bits):
        ops = self.backend
        images = ops.convert_to_tensor(images)
        bits = ops.convert_to_tensor(bits)
        dtype = backend.standardize_dtype(images.dtype)

        def posterize_float(images):
            levels = ops.cast(ops.numpy.power(2, bits), images.dtype)
            if len(ops.shape(levels)) == 0:
                levels = ops.numpy.expand_dims(levels, axis=0)
            levels = ops.numpy.expand_dims(levels, axis=(1, 2, 3))
            images = ops.numpy.floor(ops.numpy.multiply(images, levels))
            images = ops.numpy.clip(images, 0, levels - 1)
            images = ops.numpy.multiply(images, 1.0 / levels)
            return images

        def posterize_int(images):
            dtype_bits = self._num_bits_of_dtype(dtype)
            mask = ops.numpy.subtract(ops.numpy.power(2, bits), 1)
            mask = ops.numpy.multiply(
                mask, ops.numpy.power(2, dtype_bits - bits)
            )
            mask = ops.cast(mask, images.dtype)
            return images & mask

        if backend.is_float_dtype(dtype):
            images = posterize_float(images)
        else:
            images = posterize_int(images)
        return images

    def sharpen(self, images, factor, data_format=None):
        data_format = data_format or backend.image_data_format()
        if data_format == "channels_last":
            channels_axis = -1
            h_axis = -3
            w_axis = -2
        else:
            channels_axis = -3
            h_axis = -2
            w_axis = -1

        ops = self.backend
        images = ops.convert_to_tensor(images)
        factor = ops.convert_to_tensor(factor)
        channels = ops.shape(images)[channels_axis]
        original_dtype = images.dtype
        float_dtype = backend.result_type(original_dtype, float)
        max_value = self._max_value_of_dtype(original_dtype)
        if len(ops.shape(factor)) == 1:
            factor = ops.numpy.expand_dims(factor, [1, 2, 3])
        images = ops.cast(images, float_dtype)
        # [1 1 1]
        # [1 5 1]
        # [1 1 1]
        kernel = (
            ops.convert_to_tensor(
                [[1, 1, 1], [1, 5, 1], [1, 1, 1]], float_dtype
            )
            / 13.0
        )
        kernel = ops.numpy.expand_dims(kernel, axis=[-1, -2])
        kernel = ops.numpy.repeat(kernel, channels, 2)
        blurred_degenerate = ops.nn.depthwise_conv(
            images, kernel, 1, data_format=data_format
        )
        if backend.is_int_dtype(original_dtype):
            blurred_degenerate = ops.numpy.round(blurred_degenerate)

        if data_format == "channels_last":
            view = images[:, 1:-1, 1:-1, :]
            top, bottom = images[:, 0:1, 1:-1, :], images[:, -1:, 1:-1, :]
            left, right = images[:, :, 0:1, :], images[:, :, -1:, :]
        else:
            view = images[:, :, 1:-1, 1:-1]
            top, bottom = images[:, :, 0:1, 1:-1], images[:, :, -1:, 1:-1]
            left, right = images[:, :, :, 0:1], images[:, :, :, -1:]

        # We speed up blending by minimizing flops and doing in-place.
        # The 2 blend options are mathematically equivalent:
        # x + (1-r) * (y-x) = x + (1-r) * y - (1-r) * x = x * r + y * (1-r)
        view = ops.numpy.add(
            view,
            ops.numpy.multiply(
                1.0 - factor, ops.numpy.subtract(blurred_degenerate, view)
            ),
        )

        # Fix the borders of the resulting images by filling in the values of
        # the original images
        images = ops.numpy.concatenate([top, view, bottom], axis=h_axis)
        images = ops.numpy.concatenate([left, images, right], axis=w_axis)

        images = ops.numpy.clip(images, 0, max_value)
        images = ops.cast(images, original_dtype)
        return images

    def solarize(self, images, threshold):
        ops = self.backend
        images = ops.convert_to_tensor(images)
        threshold = ops.convert_to_tensor(threshold)
        if len(ops.shape(threshold)) == 0:
            threshold = ops.numpy.expand_dims(threshold, axis=0)
        threshold = ops.numpy.expand_dims(threshold, axis=(1, 2, 3))
        images = ops.numpy.where(
            images >= ops.cast(threshold, images.dtype),
            self.invert(images),
            images,
        )
        return images

    def _compute_affine_matrix(
        self,
        center_x,
        center_y,
        angle,
        translate_x,
        translate_y,
        scale,
        shear_x,
        shear_y,
        height,
        width,
    ):
        ops = self.backend
        batch_size = ops.shape(angle)[0]
        dtype = angle.dtype
        width = ops.cast(width, dtype)
        height = ops.cast(height, dtype)
        cx = center_x * width
        cy = center_y * height
        degree_to_radian_factor = math.pi / 180.0
        rot = ops.numpy.multiply(angle, degree_to_radian_factor)
        tx = -translate_x * width
        ty = -translate_y * height
        sx = ops.numpy.multiply(shear_x, degree_to_radian_factor)
        sy = ops.numpy.multiply(shear_y, degree_to_radian_factor)

        # Compute rotation & scaling & shear & translation matrix
        # cv2.getRotationMatrix2D
        alpha = ops.numpy.cos(rot) / scale
        beta = ops.numpy.sin(rot) / scale
        matrix = ops.numpy.stack(
            [
                alpha,
                # + sx (shear)
                beta + sx,
                # - cx * sx (shear) + tx (translate)
                (1.0 - alpha) * cx - beta * cy - cx * sx + tx,
                # + sy (shear)
                -beta + sy,
                alpha,
                # - cy * sy (shear) + ty (translate)
                beta * cx + (1.0 - alpha) * cy - cy * sy + ty,
                ops.numpy.zeros([batch_size], dtype),
                ops.numpy.zeros([batch_size], dtype),
                ops.numpy.ones([batch_size], dtype),
            ],
            axis=-1,
        )
        matrix = ops.numpy.reshape(matrix, [batch_size, 3, 3])
        return matrix

    def _compute_inverse_affine_matrix(
        self,
        center_x,
        center_y,
        angle,
        translate_x,
        translate_y,
        scale,
        shear_x,
        shear_y,
        height,
        width,
    ):
        # Ref: TF._geometry._get_inverse_affine_matrix
        ops = self.backend
        batch_size = ops.shape(angle)[0]
        dtype = angle.dtype
        width = ops.cast(width, dtype)
        height = ops.cast(height, dtype)

        angle = -angle
        shear_x = -shear_x
        shear_y = -shear_y

        cx = center_x * width
        cy = center_y * height
        rot = ops.numpy.multiply(angle, 1.0 / 180.0 * math.pi)
        tx = -translate_x * width
        ty = -translate_y * height
        sx = ops.numpy.multiply(shear_x, 1.0 / 180.0 * math.pi)
        sy = ops.numpy.multiply(shear_y, 1.0 / 180.0 * math.pi)

        # Cached results
        cos_sy = ops.numpy.cos(sy)
        tan_sx = ops.numpy.tan(sx)
        rot_minus_sy = rot - sy
        cx_plus_tx = cx + tx
        cy_plus_ty = cy + ty

        # Rotate Scale Shear (RSS) without scaling
        a = ops.numpy.cos(rot_minus_sy) / cos_sy
        b = -(a * tan_sx + ops.numpy.sin(rot))
        c = ops.numpy.sin(rot_minus_sy) / cos_sy
        d = ops.numpy.cos(rot) - c * tan_sx

        # Inverted rotation matrix with scale and shear
        # det([[a, b], [c, d]]) == 1, since det(rotation) = 1 and det(shear) = 1
        a0 = d * scale
        a1 = -b * scale
        b0 = -c * scale
        b1 = a * scale
        a2 = cx - a0 * cx_plus_tx - a1 * cy_plus_ty
        b2 = cy - b0 * cx_plus_tx - b1 * cy_plus_ty

        # Shape of matrix: [[batch_size], ...] -> [batch_size, 6]
        matrix = ops.numpy.stack(
            [
                a0,
                a1,
                a2,
                b0,
                b1,
                b2,
                ops.numpy.zeros([batch_size], dtype),
                ops.numpy.zeros([batch_size], dtype),
                ops.numpy.ones([batch_size], dtype),
            ],
            axis=-1,
        )
        matrix = ops.numpy.reshape(matrix, [batch_size, 3, 3])
        return matrix

    def fill_rectangles(self, images, fill_images, boxes, data_format=None):
        data_format = data_format or backend.image_data_format()
        if data_format == "channels_last":
            h_axis, w_axis, c_axis = -3, -2, -1
        else:
            c_axis, h_axis, w_axis = -3, -2, -1

        ops = self.backend
        images_shape = ops.shape(images)
        height, width = images_shape[h_axis], images_shape[w_axis]
        x0, y0, x1, y1 = boxes

        def _to_mask(start, end, mask_len):
            batch_size = ops.shape(start)[0]
            axis_indices = ops.numpy.arange(mask_len, dtype=start.dtype)
            axis_indices = ops.numpy.expand_dims(axis_indices, 0)
            axis_indices = ops.numpy.tile(axis_indices, [batch_size, 1])

            start = ops.numpy.expand_dims(start, axis=-1)
            end = ops.numpy.expand_dims(end, axis=-1)
            mask = ops.numpy.logical_and(
                ops.numpy.greater_equal(axis_indices, start),
                ops.numpy.less(axis_indices, end),
            )
            return mask

        w_masks = _to_mask(x0, x1, width)
        h_masks = _to_mask(y0, y1, height)
        w_masks = ops.numpy.expand_dims(w_masks, axis=-2)
        h_masks = ops.numpy.expand_dims(h_masks, axis=-1)
        masks = ops.numpy.logical_and(w_masks, h_masks)
        masks = ops.numpy.expand_dims(masks, axis=c_axis)
        return ops.numpy.where(masks, fill_images, images)

    def _max_value_of_dtype(self, dtype):
        dtype = backend.standardize_dtype(dtype)
        if dtype == "uint8":
            return 255
        elif dtype == "uint16":
            return 65535
        elif dtype == "uint32":
            return 4294967295
        elif dtype == "uint64":
            return 18446744073709551615
        elif dtype == "int8":
            return 127
        elif dtype == "int16":
            return 32767
        elif dtype == "int32":
            return 2147483647
        elif dtype == "int64":
            return 9223372036854775807
        else:
            return 1

    def _num_bits_of_dtype(self, dtype):
        dtype = backend.standardize_dtype(dtype)
        if dtype == "uint8":
            return 8
        elif dtype == "uint16":
            return 16
        elif dtype == "uint32":
            return 32
        elif dtype == "uint64":
            return 64
        elif dtype == "int8":
            return 7
        elif dtype == "int16":
            return 15
        elif dtype == "int32":
            return 31
        elif dtype == "int64":
            return 63
        else:
            raise ValueError(
                "_num_bits_of_dtype is only defined for integer dtypes. "
                f"Received: dtype={dtype}"
            )



import typing

import keras
from keras import backend
from keras.src.utils.backend_utils import in_tf_graph

from keras_aug._src.keras_aug_export import keras_aug_export
from keras_aug._src.layers.base.vision_random_layer import VisionRandomLayer
from keras_aug._src.utils.argument_validation import standardize_bbox_format
from keras_aug._src.utils.argument_validation import standardize_data_format
from keras_aug._src.utils.argument_validation import standardize_interpolation
from keras_aug._src.utils.argument_validation import standardize_padding_mode
from keras_aug._src.utils.argument_validation import standardize_size


@keras_aug_export(parent_path=["keras_aug.layers.vision"])
@keras.saving.register_keras_serializable(package="keras_aug")
class Mosaic(VisionRandomLayer):
    """Apply Mosaic augmentation to the provided batch of images and labels.

    The Mosaic data augmentation takes exactly 4 images from the batch and
    makes a grid. Please ensure that the batch size is 4; any additional images
    will be ignored.

    Note that `Mosaic` is meant to be used on batches of inputs, not individual
    input. The sample pairing is deterministic and done by matching consecutive
    samples in the batch, so the batch needs to be shuffled.

    Typically, `Mosaic` expects the `labels` to be one-hot-encoded format. If
    they are not, with provided `num_classes`, this layer will transform the
    `labels` into one-hot-encoded format. (e.g. `(batch_size, num_classes)`)

    Args:
        size: Desired output size of the mosaic. If `size` is an int instead of
            sequence like `(h, w)`, a square `(size, size)` is made.
        offset: The offset of the mosaic center from the top-left corner of the
            mosaic. Defaults to `(0.25, 0.75)`.
        num_classes: The number of classes in the inputs. Used for one-hot
            encoding. Can be `None` if the labels are already one-hot-encoded.
            Defaults to `None`.
        interpolation: The interpolation mode. Available values are:
            `"nearest", "bilinear"`. Defaults to `"bilinear"`.
        padding_mode: The mode of the padding. Available values:
            `"constant", "edge", "reflect", "symmetric"`. Defaults to
            `"constant"`.
        padding_value: The padding value. It only affects when
            `padding_mode="constant"`. Defaults to `0`.
        bounding_box_format: The format of the bounding boxes. If specified,
            the available values are `"xyxy", "xywh", "center_xywh", "rel_xyxy",
            "rel_xywh", "rel_center_xywh"`. Defaults to `None`.
        data_format: A string specifying the data format of the input images.
            It can be either `"channels_last"` or `"channels_first"`.
            If not specified, the value will be interpreted by
            `keras.config.image_data_format`. Defaults to `None`.
    """  # noqa: E501

    def __init__(
        self,
        size: typing.Union[typing.Sequence[int], int],
        offset: typing.Sequence[float] = (0.25, 0.75),
        num_classes: typing.Optional[int] = None,
        interpolation: str = "bilinear",
        padding_mode: str = "constant",
        padding_value: float = 0,
        bounding_box_format: typing.Optional[str] = None,
        data_format: typing.Optional[str] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        # Check
        if not isinstance(offset, typing.Sequence) or len(offset) != 2:
            raise ValueError(
                "`offset` must be a sequence with length 2. "
                f"Received: offset={offset}"
            )
        self.size = standardize_size(size)
        self.offset = tuple(offset)
        self.num_classes = int(num_classes) if num_classes is not None else None
        self.interpolation = standardize_interpolation(interpolation)
        self.padding_mode = standardize_padding_mode(padding_mode)
        self.padding_value = padding_value
        self.bounding_box_format = standardize_bbox_format(bounding_box_format)
        self.data_format = standardize_data_format(data_format)

        if self.data_format == "channels_last":
            self.h_axis, self.w_axis = -3, -2
        else:
            self.h_axis, self.w_axis = -2, -1

    def compute_output_shape(
        self,
        top_left_shape,
        top_right_shape,
        bottom_left_shape,
        bottom_right_shape,
    ):
        images_shape, segmentation_masks_shape = self._get_shape_or_spec(
            top_left_shape
        )
        images_shape = list(images_shape)
        images_shape[self.h_axis] = self.size[0]
        images_shape[self.w_axis] = self.size[1]
        if segmentation_masks_shape is not None:
            segmentation_masks_shape = list(segmentation_masks_shape)
            segmentation_masks_shape[self.h_axis] = self.size[0]
            segmentation_masks_shape[self.w_axis] = self.size[1]
        output_shape = self._set_shape(
            top_left_shape, images_shape, segmentation_masks_shape
        )

        if isinstance(output_shape, dict) and "bounding_boxes" in output_shape:
            input_keys = set(output_shape["bounding_boxes"].keys())
            extra_keys = input_keys - set(("boxes", "classes"))
            if extra_keys:
                raise KeyError(
                    "There are unsupported keys in `bounding_boxes`: "
                    f"{list(extra_keys)}. "
                    "Only `boxes` and `classes` are supported."
                )
            boxes_shape = list(output_shape["bounding_boxes"]["boxes"])
            boxes_shape[1] *= 4
            classes_shape = list(output_shape["bounding_boxes"]["classes"])
            classes_shape[1] *= 4
            output_shape["bounding_boxes"]["boxes"] = boxes_shape
            output_shape["bounding_boxes"]["classes"] = classes_shape
        return output_shape

    def get_params(self, batch_size, images=None, **kwargs):
        ops = self.backend
        random_generator = self.random_generator

        centers_x = ops.random.uniform(
            (batch_size,), self.offset[0], self.offset[1], seed=random_generator
        )
        centers_y = ops.random.uniform(
            (batch_size,), self.offset[0], self.offset[1], seed=random_generator
        )
        return dict(centers_x=centers_x, centers_y=centers_y)

    def augment_images(self, images, transformations, **kwargs):
        ops = self.backend

        cx = transformations["centers_x"]
        cy = transformations["centers_y"]
        images_shape = ops.shape(images)
        batch_size = images_shape[0]
        height, width = images_shape[self.h_axis], images_shape[self.w_axis]

        # Affine
        translate_x = cx - 0.5
        translate_y = cy - 0.5
        images = self.image_backend.affine(
            images,
            ops.numpy.zeros([batch_size]),
            translate_x,  # x
            translate_y,  # y
            ops.numpy.ones([batch_size]),
            ops.numpy.zeros([batch_size]),
            ops.numpy.zeros([batch_size]),
            interpolation="bilinear",
            padding_mode=self.padding_mode,
            padding_value=self.padding_value,
            data_format=self.data_format,
        )
        # Center crop
        offset_height = ops.numpy.floor_divide((height - self.size[0]), 2)
        offset_width = ops.numpy.floor_divide((width - self.size[1]), 2)
        images = self.image_backend.crop(
            images, offset_height, offset_width, self.size[0], self.size[1]
        )
        return images

    def augment_bounding_boxes(
        self,
        bounding_boxes,
        transformations,
        images=None,
        raw_images=None,
        **kwargs,
    ):
        if self.bounding_box_format is None:
            raise ValueError(
                f"{self.__class__.__name__} was called with bounding boxes,"
                "but no `bounding_box_format` was specified in the constructor."
            )
        ops = self.backend

        images_shape = ops.shape(raw_images)
        batch_size = images_shape[0]
        height, width = images_shape[self.h_axis], images_shape[self.w_axis]
        bounding_boxes = self.bbox_backend.convert_format(
            bounding_boxes,
            source=self.bounding_box_format,
            target="xyxy",
            height=height,
            width=width,
            dtype=self.bounding_box_dtype,
        )
        cx = transformations["centers_x"]
        cy = transformations["centers_y"]

        # Affine
        translate_x = cx - 0.5
        translate_y = cy - 0.5
        boxes = self.bbox_backend.affine(
            bounding_boxes["boxes"],
            ops.numpy.zeros([batch_size]),
            translate_x,
            translate_y,
            ops.numpy.ones([batch_size]),
            ops.numpy.zeros([batch_size]),
            ops.numpy.zeros([batch_size]),
            height,
            width,
        )
        # Center crop
        offset_height = ops.numpy.floor_divide((height - self.size[0]), 2)
        offset_width = ops.numpy.floor_divide((width - self.size[1]), 2)
        offset_height = ops.cast(offset_height, self.bounding_box_dtype)
        offset_width = ops.cast(offset_width, self.bounding_box_dtype)
        boxes = self.bbox_backend.crop(
            boxes,
            offset_height,
            offset_width,
            height=self.size[0],
            width=self.size[1],
        )

        bounding_boxes = bounding_boxes.copy()
        bounding_boxes["boxes"] = boxes
        bounding_boxes = self.bbox_backend.clip_to_images(
            bounding_boxes,
            height=self.size[0],
            width=self.size[1],
            format="xyxy",
        )
        bounding_boxes = self.bbox_backend.convert_format(
            bounding_boxes,
            source="xyxy",
            target=self.bounding_box_format,
            height=self.size[0],
            width=self.size[1],
            dtype=self.bounding_box_dtype,
        )
        return bounding_boxes

    def augment_segmentation_masks(
        self, segmentation_masks, transformations, **kwargs
    ):
        ops = self.backend
        original_dtype = segmentation_masks.dtype

        cx = transformations["centers_x"]
        cy = transformations["centers_y"]
        images_shape = ops.shape(segmentation_masks)
        batch_size = images_shape[0]
        height, width = images_shape[self.h_axis], images_shape[self.w_axis]

        # Affine
        translate_x = cx - 0.5
        translate_y = cy - 0.5
        segmentation_masks = ops.cast(
            segmentation_masks,
            backend.result_type(segmentation_masks.dtype, float),
        )
        segmentation_masks = self.image_backend.affine(
            segmentation_masks,
            ops.numpy.zeros([batch_size]),
            translate_x,  # x
            translate_y,  # y
            ops.numpy.ones([batch_size]),
            ops.numpy.zeros([batch_size]),
            ops.numpy.zeros([batch_size]),
            interpolation="nearest",
            padding_mode="constant",
            padding_value=-1,
            data_format=self.data_format,
        )
        segmentation_masks = ops.cast(segmentation_masks, original_dtype)
        # Center crop
        offset_height = ops.numpy.floor_divide((height - self.size[0]), 2)
        offset_width = ops.numpy.floor_divide((width - self.size[1]), 2)
        segmentation_masks = self.image_backend.crop(
            segmentation_masks,
            offset_height,
            offset_width,
            self.size[0],
            self.size[1],
        )
        return segmentation_masks

    def __call__(
        self, top_left, top_right, bottom_left, bottom_right, **kwargs
    ):
        if in_tf_graph():
            self._set_backend("tensorflow")
            try:
                outputs = super(VisionRandomLayer, self).__call__(
                    top_left, top_right, bottom_left, bottom_right, **kwargs
                )
            finally:
                self._reset_backend()
            return outputs
        else:
            return super(VisionRandomLayer, self).__call__(
                top_left, top_right, bottom_left, bottom_right, **kwargs
            )

    def call(self, top_left, top_right, bottom_left, bottom_right, **kwargs):
        ops = self.backend

        def format_and_cast(inputs):
            inputs, metadata = self._format_inputs(inputs)
            inputs = self._cast_inputs(inputs)
            return inputs, metadata

        top_left, metadata_tl = format_and_cast(top_left)
        top_right, metadata_tr = format_and_cast(top_right)
        bottom_left, metadata_bl = format_and_cast(bottom_left)
        bottom_right, metadata_br = format_and_cast(bottom_right)

        if not (metadata_tl == metadata_tr == metadata_bl == metadata_br):
            raise ValueError(
                "All inputs must have the same structure such as batched or as "
                "a dict. "
                f"Received: {metadata_tl}, {metadata_tr}, {metadata_bl}, "
                f"{metadata_br}"
            )

        images_shape = ops.shape(top_left[self.IMAGES])
        if len(images_shape) == 4:
            return self._format_output(
                self._batch_augment(
                    top_left, top_right, bottom_left, bottom_right
                ),
                metadata_tl,
            )
        else:
            raise ValueError(
                "Image augmentation layers are expecting inputs to be "
                "rank 3D (unbatched) or 4D (batched) tensors. "
                f"Received: images.shape={images_shape}"
            )

    def _batch_augment(
        self, top_left, top_right, bottom_left, bottom_right, **kwargs
    ):
        images, bounding_boxes, segmentation_masks = self._concate_inputs(
            top_left, top_right, bottom_left, bottom_right
        )
        raw_images = images
        labels = keypoints = custom_annotations = None
        batch_size = self.backend.shape(images)[0]

        transformations = self.get_params(
            batch_size,
            images=images,
            labels=labels,
            bounding_boxes=bounding_boxes,
            keypoints=keypoints,
            segmentation_masks=segmentation_masks,
            custom_annotations=custom_annotations,
        )

        images = self.augment_images(
            images,
            transformations=transformations,
            bounding_boxes=bounding_boxes,
            labels=labels,
        )
        result = {self.IMAGES: images}

        if bounding_boxes is not None:
            bounding_boxes = self.augment_bounding_boxes(
                bounding_boxes,
                transformations=transformations,
                labels=labels,
                images=images,
                raw_images=raw_images,
            )
            result[self.BOUNDING_BOXES] = bounding_boxes

        if segmentation_masks is not None:
            segmentation_masks = self.augment_segmentation_masks(
                segmentation_masks,
                transformations=transformations,
                labels=labels,
                bounding_boxes=bounding_boxes,
                images=images,
                raw_images=raw_images,
            )
            result[self.SEGMENTATION_MASKS] = segmentation_masks

        # Cannot preserve any additional inputs unmodified by this layer.
        if len(top_left.keys() - result.keys()):
            raise NotImplementedError
        return result

    def _concate_inputs(self, top_left, top_right, bottom_left, bottom_right):
        labels = top_left.get(self.LABELS, None)
        keypoints = top_left.get(self.KEYPOINTS, None)
        custom_annotations = top_left.get(self.CUSTOM_ANNOTATIONS, None)
        if (
            labels is not None
            or keypoints is not None
            or custom_annotations is not None
        ):
            raise NotImplementedError(
                f"{self.__class__.__name__} doesn't support augmenting "
                "labels, keypoints or custom annotations."
            )

        ops = self.backend

        # Concat images
        images_tl = top_left[self.IMAGES]
        images_tr = top_right[self.IMAGES]
        images_bl = bottom_left[self.IMAGES]
        images_br = bottom_right[self.IMAGES]
        images_top = ops.numpy.concatenate(
            [images_tl, images_tr], axis=self.w_axis
        )
        images_bottom = ops.numpy.concatenate(
            [images_bl, images_br], axis=self.w_axis
        )
        images = ops.numpy.concatenate(
            [images_top, images_bottom], axis=self.h_axis
        )

        # Concat bounding boxes
        bounding_boxes = None  # Default value
        bboxes_tl = top_left.get(self.BOUNDING_BOXES, None)
        if bboxes_tl is not None:
            if self.bounding_box_format is None:
                raise ValueError(
                    f"{self.__class__.__name__} was called with bounding boxes,"
                    "but no `bounding_box_format` was specified in the "
                    "constructor."
                )
            bboxes_tr = top_right[self.BOUNDING_BOXES]
            bboxes_bl = bottom_left[self.BOUNDING_BOXES]
            bboxes_br = bottom_right[self.BOUNDING_BOXES]
            images_shape = ops.shape(images_tl)
            height = ops.cast(
                images_shape[self.h_axis], self.bounding_box_dtype
            )
            width = ops.cast(images_shape[self.w_axis], self.bounding_box_dtype)

            def compute_bboxes(bounding_boxes, position="top_left"):
                bounding_boxes = self.bbox_backend.convert_format(
                    bounding_boxes,
                    source=self.bounding_box_format,
                    target="xyxy",
                    height=height,
                    width=width,
                    dtype=self.bounding_box_dtype,
                )
                boxes = bounding_boxes["boxes"]
                classes = bounding_boxes["classes"]
                if position == "top_left":
                    return boxes, classes
                elif position == "top_right":
                    x1, y1, x2, y2 = ops.numpy.split(boxes, 4, axis=-1)
                    x1 = x1 + width
                    x2 = x2 + width
                    boxes = ops.numpy.concatenate([x1, y1, x2, y2], axis=-1)
                    return boxes, classes
                elif position == "bottom_left":
                    x1, y1, x2, y2 = ops.numpy.split(boxes, 4, axis=-1)
                    y1 = y1 + height
                    y2 = y2 + height
                    boxes = ops.numpy.concatenate([x1, y1, x2, y2], axis=-1)
                    return boxes, classes
                elif position == "bottom_right":
                    x1, y1, x2, y2 = ops.numpy.split(boxes, 4, axis=-1)
                    x1 = x1 + width
                    y1 = y1 + height
                    x2 = x2 + width
                    y2 = y2 + height
                    boxes = ops.numpy.concatenate([x1, y1, x2, y2], axis=-1)
                    return boxes, classes
                else:
                    raise ValueError

            boxes_tl, classes_tl = compute_bboxes(bboxes_tl, "top_left")
            boxes_tr, classes_tr = compute_bboxes(bboxes_tr, "top_right")
            boxes_bl, classes_bl = compute_bboxes(bboxes_bl, "bottom_left")
            boxes_br, classes_br = compute_bboxes(bboxes_br, "bottom_right")
            boxes = ops.numpy.concatenate(
                [boxes_tl, boxes_tr, boxes_bl, boxes_br], axis=1
            )
            classes = ops.numpy.concatenate(
                [classes_tl, classes_tr, classes_bl, classes_br], axis=1
            )
            bounding_boxes = bboxes_tl.copy()
            bounding_boxes["boxes"] = boxes
            bounding_boxes["classes"] = classes
            bounding_boxes = self.bbox_backend.clip_to_images(
                bounding_boxes,
                height=height * 2,
                width=width * 2,
                format="xyxy",
            )
            bounding_boxes = self.bbox_backend.convert_format(
                bounding_boxes,
                source="xyxy",
                target=self.bounding_box_format,
                height=height * 2,
                width=width * 2,
                dtype=self.bounding_box_dtype,
            )

        # Concat segmentation masks
        segmentation_masks = None  # Default value
        seg_masks_tl = top_left.get(self.SEGMENTATION_MASKS, None)
        if seg_masks_tl is not None:
            seg_masks_tr = top_right[self.SEGMENTATION_MASKS]
            seg_masks_bl = bottom_left[self.SEGMENTATION_MASKS]
            seg_masks_br = bottom_right[self.SEGMENTATION_MASKS]
            seg_masks_top = ops.numpy.concatenate(
                [seg_masks_tl, seg_masks_tr], axis=self.w_axis
            )
            seg_masks_bottom = ops.numpy.concatenate(
                [seg_masks_bl, seg_masks_br], axis=self.w_axis
            )
            segmentation_masks = ops.numpy.concatenate(
                [seg_masks_top, seg_masks_bottom], axis=self.h_axis
            )

        return images, bounding_boxes, segmentation_masks

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "size": self.size,
                "offset": self.offset,
                "num_classes": self.num_classes,
                "interpolation": self.interpolation,
                "padding_mode": self.padding_mode,
                "padding_value": self.padding_value,
                "bounding_box_format": self.bounding_box_format,
            }
        )
        return config



import typing

import keras

from keras_aug._src.keras_aug_export import keras_aug_export
from keras_aug._src.layers.base.vision_random_layer import VisionRandomLayer
from keras_aug._src.utils.argument_validation import standardize_bbox_format
from keras_aug._src.utils.argument_validation import standardize_data_format
from keras_aug._src.utils.argument_validation import standardize_padding_mode
from keras_aug._src.utils.argument_validation import standardize_size


@keras_aug_export(parent_path=["keras_aug.layers.vision"])
@keras.saving.register_keras_serializable(package="keras_aug")
class RandomCrop(VisionRandomLayer):
    """Crop the inputs at a random location.

    If image size is smaller than output size along any edge, image is padded
    with `padding_value` and then cropped.

    Args:
        size: Desired output size of the crop. If `size` is an int instead of
            sequence like `(h, w)`, a square crop `(size, size)` is made.
        padding_mode: The mode of the padding. Available values:
            `"constant", "edge", "reflect", "symmetric"`. Defaults to
            `"constant"`.
        padding_value: The padding value. It only affects when
            `padding_mode="constant"`. Defaults to `0`.
        bounding_box_format: The format of the bounding boxes. If specified,
            the available values are `"xyxy", "xywh", "center_xywh", "rel_xyxy",
            "rel_xywh", "rel_center_xywh"`. Defaults to `None`.
        data_format: A string specifying the data format of the input images.
            It can be either `"channels_last"` or `"channels_first"`.
            If not specified, the value will be interpreted by
            `keras.config.image_data_format`. Defaults to `None`.
    """

    def __init__(
        self,
        size: typing.Union[typing.Sequence[int], int],
        padding_mode: str = "constant",
        padding_value: float = 0,
        bounding_box_format: typing.Optional[str] = None,
        data_format: typing.Optional[str] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.size = standardize_size(size)
        self.padding_mode = standardize_padding_mode(padding_mode)
        self.padding_value = padding_value
        self.bounding_box_format = standardize_bbox_format(bounding_box_format)
        self.data_format = standardize_data_format(data_format)

        if self.data_format == "channels_last":
            self.h_axis, self.w_axis = -3, -2
        else:
            self.h_axis, self.w_axis = -2, -1

    def compute_output_shape(self, input_shape):
        images_shape, segmentation_masks_shape = self._get_shape_or_spec(
            input_shape
        )
        images_shape = list(images_shape)
        images_shape[self.h_axis] = self.size[0]
        images_shape[self.w_axis] = self.size[1]
        if segmentation_masks_shape is not None:
            segmentation_masks_shape = list(segmentation_masks_shape)
            segmentation_masks_shape[self.h_axis] = self.size[0]
            segmentation_masks_shape[self.w_axis] = self.size[1]
        return self._set_shape(
            input_shape, images_shape, segmentation_masks_shape
        )

    def get_params(self, batch_size, images=None, **kwargs):
        ops = self.backend
        random_generator = self.random_generator
        images_shape = ops.shape(images)
        height, width = images_shape[self.h_axis], images_shape[self.w_axis]

        padded_height = height
        padded_width = width
        cropped_height, cropped_width = self.size

        # Pad parameters
        # self.pad_if_needed always True
        pad_top = 0
        pad_bottom = 0
        pad_left = 0
        pad_right = 0
        height_diff = ops.numpy.maximum(
            ops.numpy.subtract(cropped_height, padded_height), 0
        )
        width_diff = ops.numpy.maximum(
            ops.numpy.subtract(cropped_width, padded_width), 0
        )
        padded_height = ops.numpy.add(padded_height, 2 * height_diff)
        padded_width = ops.numpy.add(padded_width, 2 * width_diff)
        pad_top = ops.numpy.add(pad_top, height_diff)
        pad_bottom = ops.numpy.add(pad_bottom, height_diff)
        pad_left = ops.numpy.add(pad_left, width_diff)
        pad_right = ops.numpy.add(pad_right, width_diff)

        # Crop parameters
        crop_top = ops.random.randint(
            (),
            minval=0,
            maxval=padded_height - cropped_height + 1,
            seed=random_generator,
        )
        crop_left = ops.random.randint(
            (),
            minval=0,
            maxval=padded_width - cropped_width + 1,
            seed=random_generator,
        )
        return dict(
            pad_top=pad_top,
            pad_bottom=pad_bottom,
            pad_left=pad_left,
            pad_right=pad_right,
            crop_top=crop_top,
            crop_left=crop_left,
        )

    def augment_images(self, images, transformations, **kwargs):
        # Pad
        pad_top = transformations["pad_top"]
        pad_bottom = transformations["pad_bottom"]
        pad_left = transformations["pad_left"]
        pad_right = transformations["pad_right"]
        images = self.image_backend.pad(
            images,
            self.padding_mode,
            pad_top,
            pad_bottom,
            pad_left,
            pad_right,
            self.padding_value,
            self.data_format,
        )

        # Crop
        crop_top = transformations["crop_top"]
        crop_left = transformations["crop_left"]
        images = self.image_backend.crop(
            images, crop_top, crop_left, self.size[0], self.size[1]
        )
        return images

    def augment_labels(self, labels, transformations, **kwargs):
        return labels

    def augment_bounding_boxes(
        self,
        bounding_boxes,
        transformations,
        images=None,
        raw_images=None,
        **kwargs,
    ):
        if self.bounding_box_format is None:
            raise ValueError(
                f"{self.__class__.__name__} was called with bounding boxes,"
                "but no `bounding_box_format` was specified in the constructor."
            )
        ops = self.backend
        images_shape = ops.shape(raw_images)
        height = images_shape[self.h_axis]
        width = images_shape[self.w_axis]
        bounding_boxes = self.bbox_backend.convert_format(
            bounding_boxes,
            source=self.bounding_box_format,
            target="xyxy",
            height=height,
            width=width,
            dtype=self.bounding_box_dtype,
        )

        # Get pad and offset
        pad_top = ops.cast(
            transformations["pad_top"], dtype=self.bounding_box_dtype
        )
        pad_left = ops.cast(
            transformations["pad_left"], dtype=self.bounding_box_dtype
        )
        crop_top = ops.cast(
            transformations["crop_top"], dtype=self.bounding_box_dtype
        )
        crop_left = ops.cast(
            transformations["crop_left"], dtype=self.bounding_box_dtype
        )
        boxes = self.bbox_backend.pad(
            bounding_boxes["boxes"], pad_top, pad_left
        )
        boxes = self.bbox_backend.crop(
            boxes, crop_top, crop_left, height=self.size[0], width=self.size[1]
        )

        bounding_boxes = bounding_boxes.copy()
        bounding_boxes["boxes"] = boxes
        bounding_boxes = self.bbox_backend.clip_to_images(
            bounding_boxes,
            height=self.size[0],
            width=self.size[1],
            format="xyxy",
        )
        bounding_boxes = self.bbox_backend.convert_format(
            bounding_boxes,
            source="xyxy",
            target=self.bounding_box_format,
            height=self.size[0],
            width=self.size[1],
            dtype=self.bounding_box_dtype,
        )
        return bounding_boxes

    def augment_segmentation_masks(
        self, segmentation_masks, transformations, **kwargs
    ):
        # Pad
        pad_top = transformations["pad_top"]
        pad_bottom = transformations["pad_bottom"]
        pad_left = transformations["pad_left"]
        pad_right = transformations["pad_right"]
        segmentation_masks = self.image_backend.pad(
            segmentation_masks,
            "constant",
            pad_top,
            pad_bottom,
            pad_left,
            pad_right,
            -1,
            self.data_format,
        )

        # Center crop
        crop_top = transformations["crop_top"]
        crop_left = transformations["crop_left"]
        segmentation_masks = self.image_backend.crop(
            segmentation_masks,
            crop_top,
            crop_left,
            self.size[0],
            self.size[1],
        )
        return segmentation_masks

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "size": self.size,
                "padding_mode": self.padding_mode,
                "padding_value": self.padding_value,
                "bounding_box_format": self.bounding_box_format,
            }
        )
        return config



import keras
from absl.testing import parameterized
from keras import backend
from keras.src import testing

from keras_aug._src.layers.vision.trivial_augment import TrivialAugmentWide
from keras_aug._src.utils.test_utils import get_images


class TrivialAugmentWideTest(testing.TestCase, parameterized.TestCase):
    def setUp(self):
        # Defaults to channels_last
        self.data_format = backend.image_data_format()
        backend.set_image_data_format("channels_last")
        return super().setUp()

    def tearDown(self) -> None:
        backend.set_image_data_format(self.data_format)
        return super().tearDown()

    # TODO: Add correctness test

    def test_shape(self):
        # Test dynamic shape
        x = keras.KerasTensor((None, None, None, 3))
        y = TrivialAugmentWide()(x)
        self.assertEqual(y.shape, (None, None, None, 3))

        # Test static shape
        x = keras.KerasTensor((None, 32, 32, 3))
        y = TrivialAugmentWide()(x)
        self.assertEqual(y.shape, (None, 32, 32, 3))

    def test_model(self):
        # Test dynamic shape
        layer = TrivialAugmentWide()
        inputs = keras.layers.Input(shape=[None, None, 3])
        outputs = layer(inputs)
        model = keras.models.Model(inputs, outputs)
        self.assertEqual(model.output_shape, (None, None, None, 3))

        # Test static shape
        layer = TrivialAugmentWide()
        inputs = keras.layers.Input(shape=[32, 32, 3])
        outputs = layer(inputs)
        model = keras.models.Model(inputs, outputs)
        self.assertEqual(model.output_shape, (None, 32, 32, 3))

    def test_config(self):
        x = get_images("float32", "channels_last")
        layer = TrivialAugmentWide()
        y = layer(x)

        layer = TrivialAugmentWide.from_config(layer.get_config())
        y2 = layer(x)
        self.assertEqual(y.shape, y2.shape)

    def test_tf_data_compatibility(self):
        import tensorflow as tf

        layer = TrivialAugmentWide()
        x = get_images("float32", "channels_last")
        ds = tf.data.Dataset.from_tensor_slices(x).batch(2).map(layer)
        for output in ds.take(1):
            self.assertIsInstance(output, tf.Tensor)
            self.assertEqual(output.shape, (2, 32, 32, 3))



import typing

import keras
from keras import backend

from keras_aug._src.keras_aug_export import keras_aug_export
from keras_aug._src.layers.base.vision_random_layer import VisionRandomLayer
from keras_aug._src.utils.argument_validation import standardize_parameter


@keras_aug_export(parent_path=["keras_aug.layers.vision"])
@keras.saving.register_keras_serializable(package="keras_aug")
class GaussianBlur(VisionRandomLayer):
    """Blurs the input images with randomly chosen Gaussian blur kernel.

    The convolution will be using 'reflect' padding corresponding to the
    kernel size, to maintain the input shape.

    Note that due to implementation limitations, a single sampled `sigma` will
    be applied to the entire batch of images.

    Args:
        kernel_size: An int or a sequence of ints specifying the size of the
            Gaussian kernel in x and y directions. The values should be odd and
            positive numbers.
        sigma: A float or a sequence of floats specifying standard deviation to
            be used for the Gaussian kernel. If float, sigma is fixed. If a
            sequence of floats, sigma is sampled uniformly in the given range.
        data_format: A string specifying the data format of the input images.
            It can be either `"channels_last"` or `"channels_first"`.
            If not specified, the value will be interpreted by
            `keras.config.image_data_format`. Defaults to `None`.
    """

    def __init__(
        self,
        kernel_size: typing.Union[int, typing.Sequence[int]],
        sigma: typing.Union[float, typing.Sequence[float]] = (0.1, 2.0),
        data_format=None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        if isinstance(kernel_size, int):
            kernel_size = (kernel_size, kernel_size)
        if isinstance(sigma, (int, float)):
            sigma = (float(sigma), float(sigma))
        self.kernel_size = tuple(kernel_size)
        self.sigma = tuple(sigma)
        self.data_format = data_format or keras.config.image_data_format()

        if len(kernel_size) != 2:
            raise ValueError(
                "The length of `kernel_size` should be 2. "
                f"Received: kernel_size={kernel_size}"
            )
        for ks in kernel_size:
            if ks <= 0 or ks % 2 == 0:
                raise ValueError(
                    "The values of `kernel_size` should be odd and positive."
                )
        standardize_parameter(
            self.sigma,
            "sigma",
            bound=(0.0, float("inf")),
            allow_none=False,
            allow_single_number=False,
        )

    def get_params(self, batch_size, images=None, **kwargs):
        ops = self.backend
        random_generator = self.random_generator

        dtype = backend.result_type(images.dtype, float)
        sigma = ops.random.uniform(
            [1],
            self.sigma[0],
            self.sigma[1],
            dtype=dtype,
            seed=random_generator,
        )
        return sigma[0]

    def compute_output_shape(self, input_shape):
        return input_shape

    def augment_images(self, images, transformations=None, **kwargs):
        sigma = transformations
        images = self.image_backend.guassian_blur(
            images,
            self.kernel_size,
            [sigma, sigma],
            data_format=self.data_format,
        )
        return images

    def augment_labels(self, labels, transformations, **kwargs):
        return labels

    def augment_bounding_boxes(self, bounding_boxes, transformations, **kwargs):
        return bounding_boxes

    def augment_segmentation_masks(
        self, segmentation_masks, transformations, **kwargs
    ):
        return segmentation_masks

    def augment_keypoints(self, keypoints, transformations, **kwargs):
        return keypoints

    def get_config(self):
        config = super().get_config()
        config.update({"kernel_size": self.kernel_size, "sigma": self.sigma})
        return config



import ml_dtypes as ml_dtypes
import numpy as np
from keras import distribution


def get_images(dtype, data_format="channels_first", size=(32, 32)):
    # channels_first
    if dtype == "float32":
        x = np.random.uniform(0, 1, (2, 3, *size)).astype(dtype)
    elif dtype == "bfloat16":
        x = np.random.uniform(0, 1, (2, 3, *size)).astype(dtype)
    elif dtype == "uint8":
        x = np.random.uniform(0, 255, (2, 3, *size)).astype(dtype)
    elif dtype == "int8":
        x = np.random.uniform(-128, 127, (2, 3, *size)).astype(dtype)
    if data_format == "channels_last":
        x = np.transpose(x, [0, 2, 3, 1])
    return x


def uses_gpu():
    # Condition used to skip tests when using the GPU
    devices = distribution.list_devices()
    if any(d.startswith("gpu") for d in devices):
        return True
    return False



import keras

from keras_aug._src.keras_aug_export import keras_aug_export
from keras_aug._src.layers.base.vision_random_layer import VisionRandomLayer


@keras_aug_export(parent_path=["keras_aug.layers.vision"])
@keras.saving.register_keras_serializable(package="keras_aug")
class RandomInvert(VisionRandomLayer):
    """Inverts the colors of the given images.

    The equation of the inversion: `y = value_range[1] - x`.

    Args:
        p: A float specifying the probability. Defaults to `0.5`.
    """

    def __init__(self, p: float = 0.5, **kwargs):
        super().__init__(**kwargs)
        self.p = float(p)

    def compute_output_shape(self, input_shape):
        return input_shape

    def get_params(self, batch_size, images=None, **kwargs):
        ops = self.backend
        random_generator = self.random_generator
        p = ops.random.uniform([batch_size], seed=random_generator)
        return p

    def augment_images(self, images, transformations, **kwargs):
        ops = self.backend
        p = transformations

        prob = ops.numpy.expand_dims(p < self.p, axis=[1, 2, 3])
        images = ops.numpy.where(
            prob, self.image_backend.invert(images), images
        )
        return images

    def augment_labels(self, labels, transformations, **kwargs):
        return labels

    def augment_bounding_boxes(self, bounding_boxes, transformations, **kwargs):
        return bounding_boxes

    def augment_segmentation_masks(
        self, segmentation_masks, transformations, **kwargs
    ):
        return segmentation_masks

    def augment_keypoints(self, keypoints, transformations, **kwargs):
        return keypoints

    def get_config(self):
        config = super().get_config()
        config.update({"p": self.p})
        return config



import keras
import numpy as np
from absl.testing import parameterized
from keras import backend
from keras.src import testing
from keras.src.testing.test_utils import named_product

from keras_aug._src.layers.vision.resize import Resize
from keras_aug._src.utils.test_utils import get_images


class ResizeTest(testing.TestCase, parameterized.TestCase):
    pil_modes_mapping = {"nearest": 0, "bilinear": 2, "bicubic": 3}

    def setUp(self):
        # Defaults to channels_last
        self.data_format = backend.image_data_format()
        backend.set_image_data_format("channels_last")
        return super().setUp()

    def tearDown(self) -> None:
        backend.set_image_data_format(self.data_format)
        return super().tearDown()

    @parameterized.named_parameters(
        named_product(
            size=[32, (40, 50), (64, 64)],
            interpolation=["nearest", "bilinear", "bicubic"],
            antialias=[True, False],
            dtype=["float32", "uint8"],
        )
    )
    def test_correctness(self, size, interpolation, antialias, dtype):
        import torch
        import torchvision.transforms.v2.functional as TF

        if size == (40, 50) and interpolation == "nearest":
            self.skipTest("TODO: Need to investigate")
        if interpolation == "nearest" and antialias is True:
            self.skipTest("Doesn't support nearest and antialias=True")
        if interpolation == "bicubic":
            self.skipTest("TODO: Need to investigate")
        torch_interpolation = self.pil_modes_mapping[interpolation]

        if dtype == "uint8":
            atol = 2.0
            rtol = 1e-6
        else:
            atol = 1e-1
            rtol = 1e-6

        # Test channels_last
        np.random.seed(42)
        x = get_images(dtype, "channels_last")
        layer = Resize(size, interpolation, antialias, dtype=dtype)
        y = layer(x)

        ref_y = TF.resize(
            torch.tensor(np.transpose(x.copy(), [0, 3, 1, 2])),
            size,
            torch_interpolation,
            antialias=antialias,
        )
        ref_y = np.transpose(ref_y.cpu().numpy(), [0, 2, 3, 1])
        self.assertDType(y, dtype)
        self.assertAllClose(y, ref_y, atol=atol, rtol=rtol)

        # Test channels_first
        backend.set_image_data_format("channels_first")
        np.random.seed(42)
        x = get_images(dtype, "channels_first")
        layer = Resize(size, interpolation, antialias, dtype=dtype)
        y = layer(x)

        ref_y = TF.resize(
            torch.tensor(x.copy()),
            size,
            torch_interpolation,
            antialias=antialias,
        )
        ref_y = ref_y.cpu().numpy()
        self.assertDType(y, dtype)
        self.assertAllClose(y, ref_y, atol=atol, rtol=rtol)

    def test_size(self):
        # Resize the small edge
        x = get_images("float32", "channels_last", (32, 48))
        layer = Resize(10)
        y = layer(x)
        self.assertEqual(tuple(y.shape), (2, 10, 15, 3))

        # Resize the long edge
        x = get_images("float32", "channels_last", (32, 48))
        layer = Resize(10, along_long_edge=True)
        y = layer(x)
        self.assertEqual(tuple(y.shape), (2, 6, 10, 3))

        # Resize to the given size
        x = get_images("float32", "channels_last", (32, 48))
        layer = Resize((23, 45))
        y = layer(x)
        self.assertEqual(tuple(y.shape), (2, 23, 45, 3))

    def test_shape(self):
        # Test dynamic shape
        x = keras.KerasTensor((None, None, None, 3))
        y = Resize(16)(x)
        self.assertEqual(y.shape, (None, None, None, 3))

        # Test static shape
        x = keras.KerasTensor((None, 32, 32, 3))
        y = Resize(16)(x)
        self.assertEqual(y.shape, (None, 16, 16, 3))

    def test_model(self):
        # Test dynamic shape
        layer = Resize(16)
        inputs = keras.layers.Input(shape=[None, None, 3])
        outputs = layer(inputs)
        model = keras.models.Model(inputs, outputs)
        self.assertEqual(model.output_shape, (None, None, None, 3))

        # Test static shape
        layer = Resize((16, 32))
        inputs = keras.layers.Input(shape=[32, 32, 3])
        outputs = layer(inputs)
        model = keras.models.Model(inputs, outputs)
        self.assertEqual(model.output_shape, (None, 16, 32, 3))

    def test_config(self):
        x = get_images("float32", "channels_last")
        layer = Resize(16)
        y = layer(x)

        layer = Resize.from_config(layer.get_config())
        y2 = layer(x)
        self.assertAllClose(y, y2)

    def test_tf_data_compatibility(self):
        import tensorflow as tf

        layer = Resize(16)
        x = get_images("float32", "channels_last")
        ds = tf.data.Dataset.from_tensor_slices(x).batch(2).map(layer)
        for output in ds.take(1):
            self.assertIsInstance(output, tf.Tensor)
            self.assertEqual(output.shape, (2, 16, 16, 3))

    def test_augment_bounding_box(self):
        # Test full bounding boxes
        images = np.zeros([1, 20, 20, 3]).astype("float32")
        boxes = {
            "boxes": np.array(
                [
                    [[0, 0, 1, 1], [0, 0, 1, 1]],
                    [[0, 0, 1, 1], [0, 0, 1, 1]],
                ],
                "float32",
            ),
            "classes": np.array([[0, 0], [0, 0]], "float32"),
        }
        input = {"images": images, "bounding_boxes": boxes}
        layer = Resize((18, 18), bounding_box_format="rel_xyxy")

        output = layer(input)
        self.assertAllClose(output["bounding_boxes"]["boxes"], boxes["boxes"])
        self.assertAllClose(
            output["bounding_boxes"]["classes"], boxes["classes"]
        )

        # Test arbitrary bounding boxes
        images = np.zeros([1, 20, 20, 3]).astype("float32")
        boxes = {
            "boxes": np.array(
                [
                    [[3, 4, 17, 18], [10, 12, 16, 19]],
                    [[0, 0, 1, 1], [15, 12, 17, 20]],
                ],
                "float32",
            ),
            "classes": np.array([[0, 1], [2, 3]], "float32"),
        }
        input = {"images": images, "bounding_boxes": boxes}
        layer = Resize((10, 15), bounding_box_format="xyxy")

        output = layer(input)
        expected_boxes = {
            "boxes": np.array(
                [
                    [[2.25, 2, 12.75, 9], [7.5, 6, 12, 9.5]],
                    [[0, 0, 0.75, 0.5], [11.25, 6, 12.75, 10]],
                ],
                "float32",
            ),
            "classes": np.array([[0, 1], [2, 3]], "float32"),
        }
        self.assertAllClose(
            output["bounding_boxes"]["boxes"], expected_boxes["boxes"]
        )
        self.assertAllClose(
            output["bounding_boxes"]["classes"], expected_boxes["classes"]
        )

    def test_augment_segmentation_mask(self):
        num_classes = 8
        images_shape = (1, 32, 32, 3)
        masks_shape = (1, 32, 32, 1)
        images = np.random.uniform(size=images_shape).astype("float32")
        masks = np.random.randint(2, size=masks_shape) * (num_classes - 1)
        inputs = {"images": images, "segmentation_masks": masks}

        # Crop to exactly 1/2 of the size
        ref_masks = masks[:, 1::2, 1::2, :]
        layer = Resize(16)
        output = layer(inputs)
        self.assertAllClose(output["segmentation_masks"], ref_masks)



from absl.testing import parameterized
from keras import backend
from keras import ops
from keras.src import testing
from keras.src.testing.test_utils import named_product

from keras_aug._src.backend.image import ImageBackend
from keras_aug._src.utils.test_utils import get_images
from keras_aug._src.utils.test_utils import uses_gpu


class ImageBackendTest(testing.TestCase, parameterized.TestCase):
    def setUp(self):
        # Defaults to channels_last
        self.data_format = backend.image_data_format()
        backend.set_image_data_format("channels_last")
        return super().setUp()

    def tearDown(self) -> None:
        backend.set_image_data_format(self.data_format)
        return super().tearDown()

    def test_crop(self):
        image_backend = ImageBackend()

        # Test channels_last
        x = get_images("float32", "channels_last")
        y = image_backend.crop(x, top=5, left=6, height=13, width=14)
        self.assertAllClose(y, x[:, 5 : 5 + 13, 6 : 6 + 14, :])

        # Test channels_first
        backend.set_image_data_format("channels_first")
        x = get_images("float32", "channels_first")
        y = image_backend.crop(x, top=5, left=6, height=13, width=14)
        self.assertAllClose(y, x[:, :, 5 : 5 + 13, 6 : 6 + 14])

    @parameterized.named_parameters(
        named_product(mode=["constant", "reflect", "symmetric"])
    )
    def test_pad(self, mode):
        image_backend = ImageBackend()

        # Test channels_last
        x = get_images("float32", "channels_last")
        y = image_backend.pad(x, mode, 2, 3, 4, 5)

        ref_y = ops.pad(
            x,
            [[0, 0], [2, 3], [4, 5], [0, 0]],
            mode,
            0 if mode == "constant" else None,
        )
        self.assertAllClose(y, ref_y)

        # Test channels_first
        backend.set_image_data_format("channels_first")
        x = get_images("float32", "channels_first")
        y = image_backend.pad(x, mode, 2, 3, 4, 5)

        ref_y = ops.pad(
            x,
            [[0, 0], [0, 0], [2, 3], [4, 5]],
            mode,
            0 if mode == "constant" else None,
        )
        self.assertAllClose(y, ref_y)

    @parameterized.named_parameters(named_product(dtype=["float32", "uint8"]))
    def test_adjust_brightness(self, dtype):
        import torch
        import torchvision.transforms.v2.functional as TF

        image_backend = ImageBackend()
        x = get_images(dtype, "channels_first")
        y = image_backend.adjust_brightness(x, 0.5)

        ref_y = TF.adjust_brightness(torch.tensor(x), 0.5)
        ref_y = ref_y.cpu().numpy()
        self.assertAllClose(y, ref_y)
        self.assertDType(y, dtype)

    @parameterized.named_parameters(named_product(dtype=["float32", "uint8"]))
    def test_adjust_contrast(self, dtype):
        import torch
        import torchvision.transforms.v2.functional as TF

        image_backend = ImageBackend()
        x = get_images(dtype, "channels_first")
        y = image_backend.adjust_contrast(x, 0.5, "channels_first")

        ref_y = TF.adjust_contrast(torch.tensor(x), 0.5)
        ref_y = ref_y.cpu().numpy()
        self.assertAllClose(y, ref_y)
        self.assertDType(y, dtype)

    @parameterized.named_parameters(named_product(dtype=["float32", "uint8"]))
    def test_adjust_hue(self, dtype):
        import torch
        import torchvision.transforms.v2.functional as TF

        image_backend = ImageBackend()
        x = get_images(dtype, "channels_first")
        y = image_backend.adjust_hue(x, 0.5, "channels_first")

        ref_y = TF.adjust_hue(torch.tensor(x), 0.5)
        ref_y = ref_y.cpu().numpy()
        self.assertAllClose(y, ref_y)
        self.assertDType(y, dtype)

    @parameterized.named_parameters(named_product(dtype=["float32", "uint8"]))
    def test_adjust_saturation(self, dtype):
        import torch
        import torchvision.transforms.v2.functional as TF

        image_backend = ImageBackend()
        x = get_images(dtype, "channels_first")
        y = image_backend.adjust_saturation(x, 0.5, "channels_first")

        ref_y = TF.adjust_saturation(torch.tensor(x), 0.5)
        ref_y = ref_y.cpu().numpy()
        self.assertAllClose(y, ref_y)
        self.assertDType(y, dtype)

    @parameterized.named_parameters(
        named_product(
            dtype=["float32", "uint8"],
            args=[
                # angle, translate_x, translate_y, scale, shear_x, shear_y
                (15.0, 0.0, 0.0, 1.0, 0.0, 0.0),  # Test angle
                (0.0, 0.1, -0.2, 1.0, 0.0, 0.0),  # Test translate
                (0.0, 0.0, 0.0, 0.5, 0.0, 0.0),  # Test scale
                (0.0, 0.0, 0.0, 1.0, 15.0, -15.0),  # Test shear
            ],
            interpolation=["bilinear", "nearest"],
        )
    )
    def test_affine(self, dtype, args, interpolation):
        import torch
        import torchvision.transforms.v2.functional as TF

        image_backend = ImageBackend()
        x = get_images(dtype, "channels_first")

        angle, translate_x, translate_y, scale, shear_x, shear_y = args
        torch_translate_x = translate_x * 32
        torch_translate_y = translate_y * 32
        if interpolation == "bilinear":
            torch_interpolation = TF.InterpolationMode.BILINEAR
        elif interpolation == "nearest":
            torch_interpolation = TF.InterpolationMode.NEAREST

        y = image_backend.affine(
            x,
            ops.array([angle, angle]),
            ops.array([translate_x, translate_x]),
            ops.array([translate_y, translate_y]),
            ops.array([scale, scale]),
            ops.array([shear_x, shear_x]),
            ops.array([shear_y, shear_y]),
            interpolation=interpolation,
            data_format="channels_first",
        )

        ref_y = TF.affine(
            torch.tensor(x),
            angle,
            [torch_translate_x, torch_translate_y],
            scale,
            [shear_x, shear_y],
            torch_interpolation,
        )
        ref_y = ref_y.cpu().numpy()
        # TODO: Test uint8
        if dtype != "uint8":
            # TODO: Investigate these parameters
            if angle != 0.0 and interpolation == "nearest":
                pass
            elif scale != 1.0 and interpolation == "bilinear":
                pass
            elif shear_x != 0.0 or shear_y != 0.0:
                pass
            else:
                self.assertAllClose(y, ref_y, atol=0.5)
        self.assertDType(y, dtype)

    @parameterized.named_parameters(named_product(dtype=["float32", "uint8"]))
    def test_auto_contrast(self, dtype):
        import torch
        import torchvision.transforms.v2.functional as TF

        if dtype == "uint8":
            atol = 1
        else:
            atol = 1e-6

        image_backend = ImageBackend()
        x = get_images(dtype, "channels_first")
        y = image_backend.auto_contrast(x, "channels_first")

        ref_y = TF.autocontrast(torch.tensor(x))
        ref_y = ref_y.cpu().numpy()
        self.assertAllClose(y, ref_y, atol=atol)
        self.assertDType(y, dtype)

    @parameterized.named_parameters(named_product(dtype=["float32", "uint8"]))
    def test_blend(self, dtype):
        import torch
        import torchvision.transforms.v2.functional as TF

        image_backend = ImageBackend()
        x1 = get_images(dtype, "channels_first")
        x2 = get_images(dtype, "channels_first")
        y = image_backend.blend(x1, x2, 0.5)

        ref_y = TF._color._blend(torch.tensor(x1), torch.tensor(x2), 0.5)
        ref_y = ref_y.cpu().numpy()
        self.assertAllClose(y, ref_y)
        self.assertDType(y, dtype)

    @parameterized.named_parameters(named_product(dtype=["float32", "uint8"]))
    def test_equalize(self, dtype):
        import torch
        import torchvision.transforms.v2.functional as TF

        image_backend = ImageBackend()
        x = get_images(dtype, "channels_first")
        # TODO: Reduce atol
        if dtype == "float32":
            atol = 0.3
        elif dtype == "uint8":
            atol = 64
        y = image_backend.equalize(x, data_format="channels_first")

        ref_y = TF.equalize(torch.tensor(x))
        ref_y = ref_y.cpu().numpy()
        self.assertAllClose(y, ref_y, atol=atol)
        self.assertDType(y, dtype)

    @parameterized.named_parameters(named_product(dtype=["float32", "uint8"]))
    def test_guassian_blur(self, dtype):
        import torch
        import torchvision.transforms.v2.functional as TF

        if backend.backend() == "tensorflow" and not uses_gpu():
            self.skipTest("Tensorflow CPU doesn't support `guassian_blur`")
        image_backend = ImageBackend()
        x = get_images(dtype, "channels_first")
        if dtype == "float32":
            atol = 1
        elif dtype == "uint8":
            atol = 1e-6
        y = image_backend.guassian_blur(
            x, (3, 3), (0.1, 0.1), data_format="channels_first"
        )

        ref_y = TF.gaussian_blur(torch.tensor(x), (3, 3), (0.1, 0.1))
        ref_y = ref_y.cpu().numpy()
        self.assertAllClose(y, ref_y, atol=atol)
        self.assertDType(y, dtype)

    @parameterized.named_parameters(named_product(dtype=["float32", "uint8"]))
    def test_rgb_to_grayscale(self, dtype):
        import torch
        import torchvision.transforms.v2.functional as TF

        image_backend = ImageBackend()
        x = get_images(dtype, "channels_first")
        y = image_backend.rgb_to_grayscale(x, 3, "channels_first")

        ref_y = TF.rgb_to_grayscale(torch.tensor(x), 3)
        ref_y = ref_y.cpu().numpy()
        self.assertAllClose(y, ref_y)
        self.assertDType(y, dtype)

        # Test ops.image.rgb_to_grayscale
        y = ops.image.rgb_to_grayscale(x, "channels_first")

        ref_y = TF.rgb_to_grayscale(torch.tensor(x))
        ref_y = ref_y.cpu().numpy()
        self.assertAllClose(y, ref_y)
        self.assertDType(y, dtype)

    @parameterized.named_parameters(
        named_product(dtype=["float32", "uint8", "int8"])
    )
    def test_invert(self, dtype):
        import torch
        import torchvision.transforms.v2.functional as TF

        image_backend = ImageBackend()
        x = get_images(dtype, "channels_first")
        y = image_backend.invert(x)

        ref_y = TF.invert(torch.tensor(x))
        ref_y = ref_y.cpu().numpy()
        self.assertAllClose(y, ref_y)
        self.assertDType(y, dtype)

    @parameterized.named_parameters(named_product(dtype=["float32", "uint8"]))
    def test_posterize(self, dtype):
        import torch
        import torchvision.transforms.v2.functional as TF

        image_backend = ImageBackend()
        x = get_images(dtype, "channels_first")
        y = image_backend.posterize(x, bits=3)

        ref_y = TF.posterize(torch.tensor(x), bits=3)
        ref_y = ref_y.cpu().numpy()
        self.assertAllClose(y, ref_y)
        self.assertDType(y, dtype)

    @parameterized.named_parameters(named_product(dtype=["float32", "uint8"]))
    def test_sharpen(self, dtype):
        import torch
        import torchvision.transforms.v2.functional as TF

        if backend.backend() == "tensorflow" and not uses_gpu():
            self.skipTest("Tensorflow CPU doesn't support `sharpen`")
        image_backend = ImageBackend()
        x = get_images(dtype, "channels_first")
        y = image_backend.sharpen(x, 0.5, "channels_first")

        ref_y = TF.adjust_sharpness(torch.tensor(x), 0.5)
        ref_y = ref_y.cpu().numpy()
        self.assertAllClose(y, ref_y)
        self.assertDType(y, dtype)

    @parameterized.named_parameters(named_product(dtype=["float32", "uint8"]))
    def test_solarize(self, dtype):
        import torch
        import torchvision.transforms.v2.functional as TF

        image_backend = ImageBackend()
        x = get_images(dtype, "channels_first")
        if backend.is_float_dtype(dtype):
            threshold = 0.5
        elif dtype == "uint8":
            threshold = 127
        y = image_backend.solarize(x, threshold=threshold)

        ref_y = TF.solarize(torch.tensor(x), threshold=threshold)
        ref_y = ref_y.cpu().numpy()
        self.assertAllClose(y, ref_y)
        self.assertDType(y, dtype)



import typing

import keras
import numpy as np
from keras import backend

from keras_aug._src.keras_aug_export import keras_aug_export
from keras_aug._src.layers.base.vision_random_layer import VisionRandomLayer
from keras_aug._src.utils.argument_validation import standardize_bbox_format
from keras_aug._src.utils.argument_validation import standardize_interpolation
from keras_aug._src.utils.argument_validation import standardize_padding_mode


@keras_aug_export(parent_path=["keras_aug.layers.vision"])
@keras.saving.register_keras_serializable(package="keras_aug")
class RandAugment(VisionRandomLayer):
    """RandAugment data augmentation method.

    Note that due to implementation limitations, the randomness occurs in a
    batch manner.

    References:
    - [RandAugment: Practical automated data augmentation with a reduced search space](https://arxiv.org/abs/1909.13719)

    Args:
        num_ops: Number of augmentation transformations to apply sequentially.
            Defaults to `2`.
        magnitude: Magnitude for all the transformations. Defaults to `9`.
        num_magnitude_bins: The number of different magnitude values. Defaults
            to `31`.
        geometric: Whether to include geometric augmentations. This
            can be set to `False` when the inputs containing bounding boxes.
            Defaults to `True`.
        interpolation: The interpolation mode. Available values are:
            `"nearest", "bilinear"`. Defaults to `"bilinear"`.
        padding_mode: The mode of the padding. Available values:
            `"constant", "edge", "reflect", "symmetric"`. Defaults to
            `"constant"`.
        padding_value: The padding value. It only affects when
            `padding_mode="constant"`. Defaults to `0`.
        bounding_box_format: The format of the bounding boxes. If specified,
            the available values are `"xyxy", "xywh", "center_xywh", "rel_xyxy",
            "rel_xywh", "rel_center_xywh"`. Defaults to `None`.
        data_format: A string specifying the data format of the input images.
            It can be either `"channels_last"` or `"channels_first"`.
            If not specified, the value will be interpreted by
            `keras.config.image_data_format`. Defaults to `None`.
    """  # noqa: E501

    def __init__(
        self,
        num_ops: int = 2,
        magnitude: int = 9,
        num_magnitude_bins: int = 31,
        geometric: bool = True,
        interpolation: str = "bilinear",
        padding_mode: str = "constant",
        padding_value: float = 0,
        bounding_box_format: typing.Optional[str] = None,
        data_format=None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        # Check
        if magnitude >= num_magnitude_bins or magnitude < 0:
            raise ValueError(
                "`magnitude` must be a positive value and lower that "
                "`num_magnitude_bins`. "
                f"Received: magnitude={magnitude}, "
                f"num_magnitude_bins={num_magnitude_bins}"
            )

        self.num_ops = int(num_ops)
        self.magnitude = int(magnitude)
        self.num_magnitude_bins = int(num_magnitude_bins)
        self.geometric = bool(geometric)
        self.interpolation = standardize_interpolation(interpolation)
        self.padding_mode = standardize_padding_mode(padding_mode)
        self.padding_value = padding_value
        self.bounding_box_format = standardize_bbox_format(bounding_box_format)
        self.data_format = data_format or keras.config.image_data_format()

        if self.data_format == "channels_last":
            self.h_axis, self.w_axis = -3, -2
        else:
            self.h_axis, self.w_axis = -2, -1
        num_bins = self.num_magnitude_bins
        self.augmentation_space = {
            "Identity": None,
            "Brightness": np.linspace(0.0, 0.9, num_bins, dtype="float32"),
            "Color": np.linspace(0.0, 0.9, num_bins, dtype="float32"),
            "Contrast": np.linspace(0.0, 0.9, num_bins, dtype="float32"),
            "Sharpness": np.linspace(0.0, 0.9, num_bins, dtype="float32"),
            "Posterize": (
                (8 - (np.arange(num_bins) / ((num_bins - 1) / 4)))
                .round()
                .astype("int32")
            ),
            "Solarize": np.linspace(1.0, 0.0, num_bins, dtype="float32"),
            "AutoContrast": None,
            "Equalize": None,
        }
        if self.geometric:
            self.augmentation_space.update(
                {
                    "ShearX": np.degrees(
                        np.arctan(np.linspace(0.0, 0.3, num_bins))
                    ).astype("float32"),
                    "ShearY": np.degrees(
                        np.arctan(np.linspace(0.0, 0.3, num_bins))
                    ).astype("float32"),
                    "TranslateX": np.linspace(
                        0.0, 150.0 / 331.0, num_bins, dtype="float32"
                    ),
                    "TranslateY": np.linspace(
                        0.0, 150.0 / 331.0, num_bins, dtype="float32"
                    ),
                    "Rotate": np.linspace(0.0, 30.0, num_bins, dtype="float32"),
                }
            )
        p = [1.0] * len(self.augmentation_space)
        total = sum(p)
        self.p = [prob / total for prob in p]

    def compute_output_shape(self, input_shape):
        return input_shape

    def get_params(self, batch_size, images=None, **kwargs):
        ops = self.backend
        random_generator = self.random_generator

        magnitude = ops.numpy.full(
            [self.num_ops, batch_size], self.magnitude, dtype="int32"
        )
        fn_idx_p = ops.convert_to_tensor([self.p])
        fn_idx = ops.random.categorical(
            ops.numpy.log(fn_idx_p), self.num_ops, seed=random_generator
        )
        fn_idx = fn_idx[0]
        signed_p = ops.random.uniform([batch_size]) > 0.5
        signed = ops.cast(ops.numpy.where(signed_p, 1.0, -1.0), dtype="float32")
        return dict(
            magnitude=magnitude,  # shape: (self.num_ops, batch_size)
            fn_idx=fn_idx,  # shape: (self.num_ops,)
            signed=signed,  # shape: (batch_size,)
        )

    def _apply_images_transform(self, images, magnitude, idx, signed):
        ops = self.backend

        dtype = backend.standardize_dtype(images.dtype)
        batch_size = ops.shape(images)[0]

        # Build branches for ops.switch
        aug_space = self.augmentation_space
        max_value = self.image_backend._max_value_of_dtype(dtype)
        transforms = []
        for key in sorted(self.augmentation_space.keys()):
            if key == "Identity":
                transforms.append(lambda x: x)
            elif key == "Brightness":
                factor = ops.numpy.add(
                    1.0,
                    ops.numpy.multiply(
                        signed,
                        ops.numpy.take(aug_space["Brightness"], magnitude),
                    ),
                )
                transforms.append(
                    lambda x: self.image_backend.adjust_brightness(x, factor)
                )
            elif key == "Color":
                factor = ops.numpy.add(
                    1.0,
                    signed * ops.numpy.take(aug_space["Color"], magnitude),
                )
                transforms.append(
                    lambda x: self.image_backend.adjust_saturation(x, factor)
                )
            elif key == "Contrast":
                factor = ops.numpy.add(
                    1.0,
                    signed * ops.numpy.take(aug_space["Contrast"], magnitude),
                )
                transforms.append(
                    lambda x: self.image_backend.adjust_contrast(
                        x, factor, data_format=self.data_format
                    )
                )
            elif key == "Sharpness":
                factor = ops.numpy.add(
                    1.0,
                    signed * ops.numpy.take(aug_space["Sharpness"], magnitude),
                )
                transforms.append(
                    lambda x: self.image_backend.sharpen(
                        x, factor, data_format=self.data_format
                    )
                )
            elif key == "Posterize":
                bits = ops.numpy.take(aug_space["Posterize"], magnitude)
                transforms.append(
                    lambda x: self.image_backend.posterize(x, bits)
                )
            elif key == "Solarize":
                factor = ops.numpy.multiply(
                    max_value, ops.numpy.take(aug_space["Solarize"], magnitude)
                )
                transforms.append(
                    lambda x: self.image_backend.solarize(x, factor)
                )
            elif key == "AutoContrast":
                transforms.append(
                    lambda x: self.image_backend.auto_contrast(
                        x, data_format=self.data_format
                    )
                )
            elif key == "Equalize":
                transforms.append(
                    lambda x: self.image_backend.equalize(
                        x, data_format=self.data_format
                    )
                )
            elif key == "ShearX":
                shear_x = ops.numpy.multiply(
                    signed,
                    ops.numpy.full(
                        [batch_size],
                        ops.numpy.take(aug_space["ShearX"], magnitude),
                    ),
                )
                transforms.append(
                    lambda x: self.image_backend.affine(
                        x,
                        ops.numpy.zeros([batch_size]),
                        ops.numpy.zeros([batch_size]),
                        ops.numpy.zeros([batch_size]),
                        ops.numpy.ones([batch_size]),
                        shear_x,
                        ops.numpy.zeros([batch_size]),
                        interpolation=self.interpolation,
                        padding_mode=self.padding_mode,
                        padding_value=self.padding_value,
                        data_format=self.data_format,
                    )
                )
            elif key == "ShearY":
                shear_y = ops.numpy.multiply(
                    signed,
                    ops.numpy.full(
                        [batch_size],
                        ops.numpy.take(aug_space["ShearY"], magnitude),
                    ),
                )
                transforms.append(
                    lambda x: self.image_backend.affine(
                        x,
                        ops.numpy.zeros([batch_size]),
                        ops.numpy.zeros([batch_size]),
                        ops.numpy.zeros([batch_size]),
                        ops.numpy.ones([batch_size]),
                        ops.numpy.zeros([batch_size]),
                        shear_y,
                        interpolation=self.interpolation,
                        padding_mode=self.padding_mode,
                        padding_value=self.padding_value,
                        data_format=self.data_format,
                    )
                )
            elif key == "TranlateX":
                translate_x = ops.numpy.multiply(
                    signed,
                    ops.numpy.full(
                        [batch_size],
                        ops.numpy.take(aug_space["TranlateX"], magnitude),
                    ),
                )
                transforms.append(
                    lambda x: self.image_backend.affine(
                        x,
                        ops.numpy.zeros([batch_size]),
                        translate_x,
                        ops.numpy.zeros([batch_size]),
                        ops.numpy.ones([batch_size]),
                        ops.numpy.zeros([batch_size]),
                        ops.numpy.zeros([batch_size]),
                        interpolation=self.interpolation,
                        padding_mode=self.padding_mode,
                        padding_value=self.padding_value,
                        data_format=self.data_format,
                    )
                )
            elif key == "TranlateY":
                translate_y = ops.numpy.multiply(
                    signed,
                    ops.numpy.full(
                        [batch_size],
                        ops.numpy.take(aug_space["TranlateY"], magnitude),
                    ),
                )
                transforms.append(
                    lambda x: self.image_backend.affine(
                        x,
                        ops.numpy.zeros([batch_size]),
                        ops.numpy.zeros([batch_size]),
                        translate_y,
                        ops.numpy.ones([batch_size]),
                        ops.numpy.zeros([batch_size]),
                        ops.numpy.zeros([batch_size]),
                        interpolation=self.interpolation,
                        padding_mode=self.padding_mode,
                        padding_value=self.padding_value,
                        data_format=self.data_format,
                    )
                )
            elif key == "Rotate":
                angle = ops.numpy.multiply(
                    signed,
                    ops.numpy.full(
                        [batch_size],
                        ops.numpy.take(aug_space["Rotate"], magnitude),
                    ),
                )
                transforms.append(
                    lambda x: self.image_backend.affine(
                        x,
                        angle,
                        ops.numpy.zeros([batch_size]),
                        ops.numpy.zeros([batch_size]),
                        ops.numpy.ones([batch_size]),
                        ops.numpy.zeros([batch_size]),
                        ops.numpy.zeros([batch_size]),
                        interpolation=self.interpolation,
                        padding_mode=self.padding_mode,
                        padding_value=self.padding_value,
                        data_format=self.data_format,
                    )
                )
        images = ops.core.switch(idx, transforms, images)
        return images

    def augment_images(self, images, transformations, **kwargs):
        magnitude = transformations["magnitude"]
        fn_idx = transformations["fn_idx"]
        signed = transformations["signed"]
        for i in range(self.num_ops):
            idx = fn_idx[i]
            m = magnitude[i]
            images = self._apply_images_transform(images, m, idx, signed)
        return images

    def augment_labels(self, labels, transformations, **kwargs):
        return labels

    def _apply_bounding_boxes_transform(
        self, bounding_boxes, height, width, magnitude, idx, signed
    ):
        ops = self.backend

        batch_size = ops.shape(bounding_boxes["boxes"])[0]

        # Build branches for ops.switch
        aug_space = self.augmentation_space
        transforms = []
        for key in sorted(self.augmentation_space.keys()):
            if key == "Identity":
                transforms.append(lambda x: x)
            elif key == "Brightness":
                transforms.append(lambda x: x)
            elif key == "Color":
                transforms.append(lambda x: x)
            elif key == "Contrast":
                transforms.append(lambda x: x)
            elif key == "Sharpness":
                transforms.append(lambda x: x)
            elif key == "Posterize":
                transforms.append(lambda x: x)
            elif key == "Solarize":
                transforms.append(lambda x: x)
            elif key == "AutoContrast":
                transforms.append(lambda x: x)
            elif key == "Equalize":
                transforms.append(lambda x: x)
            elif key == "ShearX":
                shear_x = ops.numpy.multiply(
                    signed,
                    ops.numpy.full(
                        [batch_size],
                        ops.numpy.take(aug_space["ShearX"], magnitude),
                    ),
                )
                transforms.append(
                    lambda x: self.bbox_backend.affine(
                        x,
                        ops.numpy.zeros([batch_size]),
                        ops.numpy.zeros([batch_size]),
                        ops.numpy.zeros([batch_size]),
                        ops.numpy.ones([batch_size]),
                        shear_x,
                        ops.numpy.zeros([batch_size]),
                        height=height,
                        width=width,
                    )
                )
            elif key == "ShearY":
                shear_y = ops.numpy.multiply(
                    signed,
                    ops.numpy.full(
                        [batch_size],
                        ops.numpy.take(aug_space["ShearY"], magnitude),
                    ),
                )
                transforms.append(
                    lambda x: self.bbox_backend.affine(
                        x,
                        ops.numpy.zeros([batch_size]),
                        ops.numpy.zeros([batch_size]),
                        ops.numpy.zeros([batch_size]),
                        ops.numpy.ones([batch_size]),
                        ops.numpy.zeros([batch_size]),
                        shear_y,
                        height=height,
                        width=width,
                    )
                )
            elif key == "TranlateX":
                translate_x = ops.numpy.multiply(
                    signed,
                    ops.numpy.full(
                        [batch_size],
                        ops.numpy.take(aug_space["TranlateX"], magnitude),
                    ),
                )
                transforms.append(
                    lambda x: self.bbox_backend.affine(
                        x,
                        ops.numpy.zeros([batch_size]),
                        translate_x,
                        ops.numpy.ones([batch_size]),
                        ops.numpy.zeros([batch_size]),
                        ops.numpy.zeros([batch_size]),
                        height=height,
                        width=width,
                    )
                )
            elif key == "TranlateY":
                translate_y = ops.numpy.multiply(
                    signed,
                    ops.numpy.full(
                        [batch_size],
                        ops.numpy.take(aug_space["TranlateY"], magnitude),
                    ),
                )
                transforms.append(
                    lambda x: self.bbox_backend.affine(
                        x,
                        ops.numpy.zeros([batch_size]),
                        ops.numpy.zeros([batch_size]),
                        translate_y,
                        ops.numpy.ones([batch_size]),
                        ops.numpy.zeros([batch_size]),
                        ops.numpy.zeros([batch_size]),
                        height=height,
                        width=width,
                    )
                )
            elif key == "Rotate":
                angle = ops.numpy.multiply(
                    signed,
                    ops.numpy.full(
                        [batch_size],
                        ops.numpy.take(aug_space["Rotate"], magnitude),
                    ),
                )
                transforms.append(
                    lambda x: self.bbox_backend.affine(
                        x,
                        angle,
                        ops.numpy.zeros([batch_size]),
                        ops.numpy.zeros([batch_size]),
                        ops.numpy.ones([batch_size]),
                        ops.numpy.zeros([batch_size]),
                        ops.numpy.zeros([batch_size]),
                        height=height,
                        width=width,
                    )
                )
        boxes = ops.core.switch(idx, transforms, bounding_boxes["boxes"])
        bounding_boxes = bounding_boxes.copy()
        bounding_boxes["boxes"] = boxes
        return bounding_boxes

    def augment_bounding_boxes(
        self,
        bounding_boxes,
        transformations,
        images=None,
        raw_images=None,
        **kwargs,
    ):
        if self.bounding_box_format is None:
            raise ValueError(
                f"{self.__class__.__name__} was called with bounding boxes,"
                "but no `bounding_box_format` was specified in the constructor."
            )
        ops = self.backend

        magnitude = transformations["magnitude"]
        fn_idx = transformations["fn_idx"]
        signed = transformations["signed"]
        images_shape = ops.shape(raw_images)
        height = images_shape[self.h_axis]
        width = images_shape[self.w_axis]
        bounding_boxes = self.bbox_backend.convert_format(
            bounding_boxes,
            source=self.bounding_box_format,
            target="xyxy",
            height=height,
            width=width,
            dtype=self.bounding_box_dtype,
        )
        for i in range(self.num_ops):
            idx = fn_idx[i]
            m = magnitude[i]
            bounding_boxes = self._apply_bounding_boxes_transform(
                bounding_boxes, height, width, m, idx, signed
            )
        bounding_boxes = self.bbox_backend.clip_to_images(
            bounding_boxes,
            height=height,
            width=width,
            format="xyxy",
        )
        bounding_boxes = self.bbox_backend.convert_format(
            bounding_boxes,
            source="xyxy",
            target=self.bounding_box_format,
            height=height,
            width=width,
            dtype=self.bounding_box_dtype,
        )
        return bounding_boxes

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "num_ops": self.num_ops,
                "magnitude": self.magnitude,
                "num_magnitude_bins": self.num_magnitude_bins,
                "geometric": self.geometric,
                "interpolation": self.interpolation,
                "padding_mode": self.padding_mode,
                "padding_value": self.padding_value,
            }
        )
        return config
